"""# PySpark.

The `daft.pyspark` module provides a way to create a PySpark session that can be run locally or backed by a ray cluster.

This serves as a way to run the daft query engine, but with a spark compatible API.


## Example

```py
from daft.pyspark import SparkSession
from pyspark.sql.functions import col

# create a local spark session
spark = SparkSession.builder.local().getOrCreate()

# alternatively, connect to a ray cluster

spark = SparkSession.builder.remote("ray://<HEAD_IP>:6379").getOrCreate()
# you can use `ray get-head-ip <cluster_config.yaml>` to get the head ip!
# use spark as you would with the native spark library, but with a daft backend!

spark.createDataFrame([{"hello": "world"}]).select(col("hello")).show()

# stop the spark session
spark.stop()
```
"""

from typing import Any
from daft.daft import ConnectionHandle, connect_start
from pyspark.sql import SparkSession as PySparkSession


class Builder:
    def __init__(self) -> None:
        self._builder = PySparkSession.builder

    def local(self) -> "Builder":
        self._connection = connect_start()
        url = f"sc://0.0.0.0:{self._connection.port()}"
        self._builder = PySparkSession.builder.remote(url)
        return self

    def remote(self, url: str) -> "Builder":
        if url.startswith("ray://"):
            import daft

            if url.startswith("ray://localhost") or url.startswith("ray://127.0.0.1"):
                daft.context.set_runner_ray(noop_if_initialized=True)
            else:
                daft.context.set_runner_ray(address=url, noop_if_initialized=True)
            self._connection = connect_start()
            url = f"sc://0.0.0.0:{self._connection.port()}"
            self._builder = PySparkSession.builder.remote(url)
            return self
        else:
            self._builder = PySparkSession.builder.remote(url)
            return self

    def getOrCreate(self) -> "SparkSession":
        return SparkSession(self._builder.getOrCreate(), self._connection)

    def __getattr__(self, name: str) -> Any:
        attr = getattr(self._builder, name)
        if callable(attr):

            def wrapped(*args: Any, **kwargs: Any) -> Any:
                result = attr(*args, **kwargs)
                # If result is the original builder, return self instead
                return self if result == self._builder else result

            return wrapped
        return attr

    __doc__ = property(lambda self: self._spark_session.__doc__)  # type: ignore


class SparkSession:
    builder = Builder()

    def __init__(self, spark_session: PySparkSession, connection: ConnectionHandle) -> None:
        self._spark_session = spark_session
        self._connection = connection

    def __repr__(self) -> str:
        return self._spark_session.__repr__()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._spark_session, name)

    def stop(self) -> None:
        self._spark_session.stop()
        self._connection.shutdown()

    __doc__ = property(lambda self: self._spark_session.__doc__)  # type: ignore



from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft.io.object_store_options import io_config_to_storage_options
from daft.logical.schema import Schema
from tests.utils import assert_pyarrow_tables_equal

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(
    PYARROW_LOWER_BOUND_SKIP,
    reason="deltalake not supported on older versions of pyarrow",
)


def test_deltalake_read_basic(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    deltalake.write_deltalake(path, base_table)
    df = daft.read_deltalake(str(path))
    expected_schema = Schema.from_pyarrow_schema(deltalake.DeltaTable(path).schema().to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(df.to_arrow(), base_table)


def test_deltalake_read_full(deltalake_table):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, parts = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(df.to_arrow().sort_by("part_idx"), pa.concat_tables(parts).sort_by("part_idx"))


def test_deltalake_read_show(deltalake_table):
    path, catalog_table, io_config, _ = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    df.show()


def test_deltalake_read_row_group_splits(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"

    # Force 2 rowgroups
    deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)

    # Force file splitting
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=1,
        scan_tasks_max_size_bytes=100,
    ):
        df = daft.read_deltalake(str(path))
        df.collect()
        assert len(df) == 3, "Length of non-materialized data when read through deltalake should be correct"


def test_deltalake_read_row_group_splits_with_filter(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"

    # Force 2 rowgroups
    deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)

    # Force file splitting
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=1,
        scan_tasks_max_size_bytes=100,
    ):
        df = daft.read_deltalake(str(path))
        df = df.where(df["a"] > 1)
        df.collect()
        assert len(df) == 2, "Length of non-materialized data when read through deltalake should be correct"


def test_deltalake_read_row_group_splits_with_limit(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"

    # Force 2 rowgroups
    deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)

    # Force file splitting
    with daft.execution_config_ctx(
        scan_tasks_min_size_bytes=1,
        scan_tasks_max_size_bytes=100,
    ):
        df = daft.read_deltalake(str(path))
        df = df.limit(2)
        df.collect()
        assert len(df) == 2, "Length of non-materialized data when read through deltalake should be correct"


def test_deltalake_read_versioned(tmp_path, base_table):
    deltalake = pytest.importorskip("deltalake")
    path = tmp_path / "some_table"
    deltalake.write_deltalake(path, base_table)

    updated_columns = base_table.columns + [pa.array(["x", "y", "z"])]
    updated_column_names = base_table.column_names + ["new_column"]
    updated_table = pa.Table.from_arrays(updated_columns, names=updated_column_names)
    deltalake.write_deltalake(path, updated_table, mode="overwrite", schema_mode="overwrite")

    for version in [None, 1]:
        df = daft.read_deltalake(str(path), version=version)
        expected_schema = Schema.from_pyarrow_schema(deltalake.DeltaTable(path).schema().to_pyarrow())
        assert df.schema() == expected_schema
        assert_pyarrow_tables_equal(df.to_arrow(), updated_table)

    df = daft.read_deltalake(str(path), version=0)
    expected_schema = Schema.from_pyarrow_schema(deltalake.DeltaTable(path, version=0).schema().to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(df.to_arrow(), base_table)



import builtins
import datetime
from collections.abc import AsyncIterator, Iterator
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar

from daft.dataframe.display import MermaidOptions
from daft.execution import physical_plan
from daft.io import DataSink
from daft.io.scan import ScanOperator
from daft.io.sink import WriteResultType
from daft.runners.flotilla import RaySwordfishActorHandle
from daft.runners.partitioning import PartitionCacheEntry, PartitionT
from daft.sql.sql_connection import SQLConnection
from daft.udf import UDF, BoundUDFArgs, InitArgsType, UninitializedUdf

if TYPE_CHECKING:
    import pyarrow as pa
    import ray
    from pyiceberg.schema import Schema as IcebergSchema
    from pyiceberg.table import TableProperties as IcebergTableProperties

    from daft.catalog import Catalog, Table
    from daft.expressions.visitor import ExpressionVisitor
    from daft.runners.runner import Runner

R = TypeVar("R")

class ImageMode(Enum):
    """Supported image modes for Daft's image type.

    Warning:
        Currently, only the 8-bit modes (L, LA, RGB, RGBA) can be stored in a DataFrame.
        If your binary image data includes other modes, use the `mode` argument
        in `image.decode` to convert the images to a supported mode.
    """

    #: 8-bit grayscale
    L = 1

    #: 8-bit grayscale + alpha
    LA = 2

    #: 8-bit RGB
    RGB = 3

    #: 8-bit RGB + alpha
    RGBA = 4

    #: 16-bit grayscale
    L16 = 5

    #: 16-bit grayscale + alpha
    LA16 = 6

    #: 16-bit RGB
    RGB16 = 7

    #: 16-bit RGB + alpha
    RGBA16 = 8

    #: 32-bit floating RGB
    RGB32F = 9

    #: 32-bit floating RGB + alpha
    RGBA32F = 10

    @staticmethod
    def from_mode_string(mode: str) -> ImageMode:
        """Create an ImageMode from its string representation.

        Args:
            mode: String representation of the mode. This is the same as the enum
                attribute name, e.g. ``ImageMode.from_mode_string("RGB")`` would
                return ``ImageMode.RGB``.
        """
        ...

class PyWindowBoundary:
    """Represents a window frame boundary in window functions."""

    @staticmethod
    def unbounded_preceding() -> PyWindowBoundary: ...
    @staticmethod
    def unbounded_following() -> PyWindowBoundary: ...
    @staticmethod
    def offset(n: int) -> PyWindowBoundary: ...
    @staticmethod
    def range_offset(n: PyExpr) -> PyWindowBoundary: ...

class WindowFrameType(Enum):
    """Represents the type of window frame (ROWS or RANGE)."""

    Rows = 1
    Range = 2

class WindowFrame:
    """Represents a window frame specification."""

    def __init__(
        self,
        start: PyWindowBoundary,
        end: PyWindowBoundary,
    ) -> None: ...

class WindowSpec:
    """Represents a window specification for window functions."""

    @staticmethod
    def new() -> WindowSpec: ...
    def with_partition_by(self, exprs: list[PyExpr]) -> WindowSpec: ...
    def with_order_by(self, exprs: list[PyExpr], ascending: list[bool]) -> WindowSpec: ...
    def with_frame(self, frame: WindowFrame) -> WindowSpec: ...
    def with_min_periods(self, min_periods: int) -> WindowSpec: ...

class ImageFormat(Enum):
    """Supported image formats for Daft's image I/O."""

    PNG = 1
    JPEG = 2
    TIFF = 3
    GIF = 4
    BMP = 5

    @staticmethod
    def from_format_string(mode: str) -> ImageFormat:
        """Create an ImageFormat from its string representation."""
        ...

class JoinType(Enum):
    """Type of a join operation."""

    Inner = 1
    Left = 2
    Right = 3
    Outer = 4
    Semi = 5
    Anti = 6

    @staticmethod
    def from_join_type_str(join_type: str) -> JoinType:
        """Create a JoinType from its string representation.

        Args:
            join_type: String representation of the join type. This is the same as the enum
                attribute name (but snake-case), e.g. ``JoinType.from_join_type_str("inner")`` would
                return ``JoinType.Inner``.
        """
        ...

class JoinStrategy(Enum):
    """Join strategy (algorithm) to use."""

    Hash = 1
    SortMerge = 2
    Broadcast = 3

    @staticmethod
    def from_join_strategy_str(join_strategy: str) -> JoinStrategy:
        """Create a JoinStrategy from its string representation.

        Args:
            join_strategy: String representation of the join strategy. This is the same as the enum
                attribute name (but snake-case), e.g. ``JoinType.from_join_strategy_str("sort_merge")`` would
                return ``JoinStrategy.SortMerge``.
        """
        ...

class JoinSide(Enum):
    Left = 1
    Right = 2

class CountMode(Enum):
    """Supported count modes for Daft's count aggregation.

    | All   - Count both non-null and null values.
    | Valid - Count only valid values.
    | Null  - Count only null values.
    """

    All = 1
    Valid = 2
    Null = 3

    @staticmethod
    def from_count_mode_str(count_mode: str) -> CountMode:
        """Create a CountMode from its string representation.

        Args:
            count_mode: String representation of the count mode , e.g. "all", "valid", or "null".
        """
        ...

class ResourceRequest:
    """Resource request for a query fragment task."""

    num_cpus: float | None
    num_gpus: float | None
    memory_bytes: int | None

    def __init__(
        self,
        num_cpus: float | None = None,
        num_gpus: float | None = None,
        memory_bytes: int | None = None,
    ): ...
    @staticmethod
    def max_resources(resource_requests: list[ResourceRequest]) -> ResourceRequest:
        """Take a field-wise max of the list of resource requests."""
        ...

    def with_num_cpus(self, num_cpus: float | None) -> ResourceRequest: ...
    def with_num_gpus(self, num_gpus: float | None) -> ResourceRequest: ...
    def with_memory_bytes(self, memory_bytes: int | None) -> ResourceRequest: ...
    def __mul__(self, factor: float) -> ResourceRequest: ...
    def __add__(self, other: ResourceRequest) -> ResourceRequest: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: ResourceRequest) -> bool: ...  # type: ignore[override]
    def __ne__(self, other: ResourceRequest) -> bool: ...  # type: ignore[override]

class FileFormat(Enum):
    """Format of a file, e.g. Parquet, CSV, and JSON."""

    Parquet = 1
    Csv = 2
    Json = 3

    def ext(self) -> str: ...

class WriteMode(Enum):
    """Mode for writing data to a file."""

    Overwrite = 1
    OverwritePartitions = 2
    Append = 3

    @staticmethod
    def from_str(mode: str) -> WriteMode: ...

class ParquetSourceConfig:
    """Configuration of a Parquet data source."""

    coerce_int96_timestamp_unit: PyTimeUnit | None
    field_id_mapping: dict[int, PyField] | None
    row_groups: list[list[int]] | None
    chunk_size: int | None

    def __init__(
        self,
        coerce_int96_timestamp_unit: PyTimeUnit | None = None,
        field_id_mapping: dict[int, PyField] | None = None,
        row_groups: list[list[int]] | None = None,
        chunk_size: int | None = None,
    ): ...

class CsvSourceConfig:
    """Configuration of a CSV data source."""

    delimiter: str | None
    has_headers: bool
    double_quote: bool
    quote: str | None
    escape_char: str | None
    comment: str | None
    allow_variable_columns: bool
    buffer_size: int | None
    chunk_size: int | None

    def __init__(
        self,
        has_headers: bool,
        double_quote: bool,
        allow_variable_columns: bool,
        delimiter: str | None,
        quote: str | None,
        escape_char: str | None,
        comment: str | None,
        buffer_size: int | None = None,
        chunk_size: int | None = None,
    ): ...

class JsonSourceConfig:
    """Configuration of a JSON data source."""

    buffer_size: int | None
    chunk_size: int | None

    def __init__(
        self,
        buffer_size: int | None = None,
        chunk_size: int | None = None,
    ): ...

class WarcSourceConfig:
    """Configuration of a Warc data source."""
    def __init__(self) -> None: ...

class DatabaseSourceConfig:
    """Configuration of a database data source."""

    sql: str
    conn: SQLConnection

    def __init__(self, sql: str, conn_factory: SQLConnection): ...

class FileFormatConfig:
    """Configuration for parsing a particular file format (Parquet, CSV, JSON)."""

    config: ParquetSourceConfig | CsvSourceConfig | JsonSourceConfig | DatabaseSourceConfig | WarcSourceConfig

    @staticmethod
    def from_parquet_config(config: ParquetSourceConfig) -> FileFormatConfig:
        """Create a Parquet file format config."""
        ...

    @staticmethod
    def from_csv_config(config: CsvSourceConfig) -> FileFormatConfig:
        """Create a CSV file format config."""
        ...

    @staticmethod
    def from_json_config(config: JsonSourceConfig) -> FileFormatConfig:
        """Create a JSON file format config."""
        ...

    @staticmethod
    def from_warc_config(config: WarcSourceConfig) -> FileFormatConfig:
        """Create a WARC file format config."""
        ...

    @staticmethod
    def from_database_config(config: DatabaseSourceConfig) -> FileFormatConfig:
        """Create a database file format config."""
        ...

    def file_format(self) -> FileFormat:
        """Get the file format for this config."""
        ...

    def __eq__(self, other: FileFormatConfig) -> bool: ...  # type: ignore[override]
    def __ne__(self, other: FileFormatConfig) -> bool: ...  # type: ignore[override]

class CsvConvertOptions:
    """Options for converting CSV data to Daft data."""

    limit: int | None
    include_columns: list[str] | None
    column_names: list[str] | None
    schema: PySchema | None
    predicate: PyExpr | None

    def __init__(
        self,
        limit: int | None = None,
        include_columns: list[str] | None = None,
        column_names: list[str] | None = None,
        schema: PySchema | None = None,
        predicate: PyExpr | None = None,
    ): ...

class CsvParseOptions:
    """Options for parsing CSV files."""

    has_header: bool
    delimiter: str | None
    double_quote: bool
    quote: str | None
    allow_variable_columns: bool
    escape_char: str | None
    comment: str | None

    def __init__(
        self,
        has_header: bool = True,
        delimiter: str | None = None,
        double_quote: bool = True,
        quote: str | None = None,
        allow_variable_columns: bool = False,
        escape_char: str | None = None,
        comment: str | None = None,
    ): ...

class CsvReadOptions:
    """Options for reading CSV files."""

    buffer_size: int | None
    chunk_size: int | None

    def __init__(
        self,
        buffer_size: int | None = None,
        chunk_size: int | None = None,
    ): ...

class JsonConvertOptions:
    """Options for converting JSON data to Daft data."""

    limit: int | None
    include_columns: list[str] | None
    schema: PySchema | None

    def __init__(
        self,
        limit: int | None = None,
        include_columns: list[str] | None = None,
        schema: PySchema | None = None,
    ): ...

class JsonParseOptions:
    """Options for parsing JSON files."""

class JsonReadOptions:
    """Options for reading JSON files."""

    buffer_size: int | None
    chunk_size: int | None

    def __init__(
        self,
        buffer_size: int | None = None,
        chunk_size: int | None = None,
    ): ...

class FileInfo:
    """Metadata for a single file."""

    file_path: str
    file_size: int | None
    num_rows: int | None

class FileInfos:
    """Metadata for a collection of files."""

    file_paths: list[str]
    file_sizes: list[int | None]
    num_rows: list[int | None]

    @staticmethod
    def from_infos(file_paths: list[str], file_sizes: list[int | None], num_rows: list[int | None]) -> FileInfos: ...
    def extend(self, new_infos: FileInfos) -> FileInfos:
        """Concatenate two FileInfos together."""
        ...

    def __getitem__(self, idx: int) -> FileInfo: ...
    def __len__(self) -> int: ...

class HTTPConfig:
    """I/O configuration for accessing HTTP systems."""

    bearer_token: str | None

    def __init__(self, bearer_token: str | None = None): ...

class S3Config:
    """I/O configuration for accessing an S3-compatible system."""

    region_name: str | None
    endpoint_url: str | None
    key_id: str | None
    session_token: str | None
    access_key: str | None
    credentials_provider: Callable[[], S3Credentials] | None
    max_connections: int
    retry_initial_backoff_ms: int
    connect_timeout_ms: int
    read_timeout_ms: int
    num_tries: int
    retry_mode: str | None
    anonymous: bool
    use_ssl: bool
    verify_ssl: bool
    check_hostname_ssl: bool
    requester_pays: bool | None
    force_virtual_addressing: bool | None
    profile_name: str | None

    def __init__(
        self,
        region_name: str | None = None,
        endpoint_url: str | None = None,
        key_id: str | None = None,
        session_token: str | None = None,
        access_key: str | None = None,
        credentials_provider: Callable[[], S3Credentials] | None = None,
        buffer_time: int | None = None,
        max_connections: int | None = None,
        retry_initial_backoff_ms: int | None = None,
        connect_timeout_ms: int | None = None,
        read_timeout_ms: int | None = None,
        num_tries: int | None = None,
        retry_mode: str | None = None,
        anonymous: bool | None = None,
        use_ssl: bool | None = None,
        verify_ssl: bool | None = None,
        check_hostname_ssl: bool | None = None,
        requester_pays: bool | None = None,
        force_virtual_addressing: bool | None = None,
        profile_name: str | None = None,
    ): ...
    def replace(
        self,
        region_name: str | None = None,
        endpoint_url: str | None = None,
        key_id: str | None = None,
        session_token: str | None = None,
        access_key: str | None = None,
        credentials_provider: Callable[[], S3Credentials] | None = None,
        max_connections: int | None = None,
        retry_initial_backoff_ms: int | None = None,
        connect_timeout_ms: int | None = None,
        read_timeout_ms: int | None = None,
        num_tries: int | None = None,
        retry_mode: str | None = None,
        anonymous: bool | None = None,
        use_ssl: bool | None = None,
        verify_ssl: bool | None = None,
        check_hostname_ssl: bool | None = None,
        requester_pays: bool | None = None,
        force_virtual_addressing: bool | None = None,
        profile_name: str | None = None,
    ) -> S3Config:
        """Replaces values if provided, returning a new S3Config."""
        ...

    @staticmethod
    def from_env() -> S3Config:
        """Creates an S3Config, retrieving credentials and configurations from the current environment."""
        ...

    def provide_cached_credentials(self) -> S3Credentials | None:
        """Wrapper around call to `S3Config.credentials_provider` to cache credentials until expiry."""
        ...

class S3Credentials:
    key_id: str
    access_key: str
    session_token: str | None
    expiry: datetime.datetime | None

    def __init__(
        self,
        key_id: str,
        access_key: str,
        session_token: str | None = None,
        expiry: datetime.datetime | None = None,
    ): ...

class AzureConfig:
    """I/O configuration for accessing Azure Blob Storage."""

    storage_account: str | None
    access_key: str | None
    sas_token: str | None
    bearer_token: str | None
    tenant_id: str | None
    client_id: str | None
    client_secret: str | None
    use_fabric_endpoint: bool | None
    anonymous: bool | None
    endpoint_url: str | None = None
    use_ssl: bool | None = None

    def __init__(
        self,
        storage_account: str | None = None,
        access_key: str | None = None,
        sas_token: str | None = None,
        bearer_token: str | None = None,
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        use_fabric_endpoint: bool | None = None,
        anonymous: bool | None = None,
        endpoint_url: str | None = None,
        use_ssl: bool | None = None,
    ): ...
    def replace(
        self,
        storage_account: str | None = None,
        access_key: str | None = None,
        sas_token: str | None = None,
        bearer_token: str | None = None,
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        use_fabric_endpoint: bool | None = None,
        anonymous: bool | None = None,
        endpoint_url: str | None = None,
        use_ssl: bool | None = None,
    ) -> AzureConfig:
        """Replaces values if provided, returning a new AzureConfig."""
        ...

class GCSConfig:
    """I/O configuration for accessing Google Cloud Storage."""

    project_id: str | None
    credentials: str | None
    token: str | None
    anonymous: bool
    max_connections: int
    retry_initial_backoff_ms: int
    connect_timeout_ms: int
    read_timeout_ms: int
    num_tries: int

    def __init__(
        self,
        project_id: str | None = None,
        credentials: str | None = None,
        token: str | None = None,
        anonymous: bool | None = None,
        max_connections: int | None = None,
        retry_initial_backoff_ms: int | None = None,
        connect_timeout_ms: int | None = None,
        read_timeout_ms: int | None = None,
        num_tries: int | None = None,
    ): ...
    def replace(
        self,
        project_id: str | None = None,
        credentials: str | None = None,
        token: str | None = None,
        anonymous: bool | None = None,
        max_connections: int | None = None,
        retry_initial_backoff_ms: int | None = None,
        connect_timeout_ms: int | None = None,
        read_timeout_ms: int | None = None,
        num_tries: int | None = None,
    ) -> GCSConfig:
        """Replaces values if provided, returning a new GCSConfig."""
        ...

class UnityConfig:
    """I/O configuration for Unity Catalog volumes."""

    endpoint: str | None
    token: str | None

    def __init__(
        self,
        endpoint: str | None,
        token: str | None,
    ): ...
    def replace(
        self,
        endpoint: str | None,
        token: str | None,
    ) -> UnityConfig:
        """Replaces values if provided, returning a new UnityConfig."""
        ...

class IOConfig:
    """Configuration for the native I/O layer, e.g. credentials for accessing cloud storage systems."""

    s3: S3Config
    azure: AzureConfig
    gcs: GCSConfig
    http: HTTPConfig
    unity: UnityConfig

    def __init__(
        self,
        s3: S3Config | None = None,
        azure: AzureConfig | None = None,
        gcs: GCSConfig | None = None,
        http: HTTPConfig | None = None,
        unity: UnityConfig | None = None,
    ): ...
    def replace(
        self,
        s3: S3Config | None = None,
        azure: AzureConfig | None = None,
        gcs: GCSConfig | None = None,
        http: HTTPConfig | None = None,
        unity: UnityConfig | None = None,
    ) -> IOConfig:
        """Replaces values if provided, returning a new IOConfig."""
        ...

class StorageConfig:
    """Configuration for interacting with a particular storage backend."""

    # Whether or not to use a multithreaded tokio runtime for processing I/O
    multithreaded_io: bool
    io_config: IOConfig

    def __init__(self, multithreaded_io: bool, io_config: IOConfig | None): ...

class ScanTask:
    """A batch of scan tasks for reading data from an external source."""

    def num_rows(self) -> int:
        """Get number of rows that will be scanned by this ScanTask."""
        ...

    def estimate_in_memory_size_bytes(self, cfg: PyDaftExecutionConfig) -> int:
        """Estimate the In Memory Size of this ScanTask."""
        ...

    @staticmethod
    def catalog_scan_task(
        file: str,
        file_format: FileFormatConfig,
        schema: PySchema,
        storage_config: StorageConfig,
        num_rows: int | None,
        size_bytes: int | None,
        iceberg_delete_files: list[str] | None,
        pushdowns: PyPushdowns | None,
        partition_values: PyRecordBatch | None,
        stats: PyRecordBatch | None,
    ) -> ScanTask | None:
        """Create a Catalog Scan Task."""
        ...

    @staticmethod
    def sql_scan_task(
        url: str,
        file_format: FileFormatConfig,
        schema: PySchema,
        storage_config: StorageConfig,
        num_rows: int | None,
        size_bytes: int | None,
        pushdowns: PyPushdowns | None,
        stats: PyRecordBatch | None,
    ) -> ScanTask:
        """Create a SQL Scan Task."""
        ...

    @staticmethod
    def python_factory_func_scan_task(
        module: str,
        func_name: str,
        func_args: tuple[Any, ...],
        schema: PySchema,
        num_rows: int | None,
        size_bytes: int | None,
        pushdowns: PyPushdowns | None,
        stats: PyRecordBatch | None,
    ) -> ScanTask:
        """Create a Python factory function Scan Task."""
        ...

class ScanOperatorHandle:
    """A handle to a scan operator."""

    @staticmethod
    def anonymous_scan(
        files: list[str],
        schema: PySchema,
        file_format_config: FileFormatConfig,
        storage_config: StorageConfig,
    ) -> ScanOperatorHandle: ...
    @staticmethod
    def glob_scan(
        glob_path: list[str],
        file_format_config: FileFormatConfig,
        storage_config: StorageConfig,
        hive_partitioning: bool,
        infer_schema: bool,
        schema: PySchema | None = None,
        file_path_column: str | None = None,
    ) -> ScanOperatorHandle: ...
    @staticmethod
    def from_python_scan_operator(operator: ScanOperator) -> ScanOperatorHandle: ...

def logical_plan_table_scan(scan_operator: ScanOperatorHandle) -> LogicalPlanBuilder: ...

class PyPartitionField:
    """Partitioning Field of a Scan Source such as Hive or Iceberg."""

    field: PyField
    source_field: PyField | None = None
    transform: PyPartitionTransform | None = None

    def __init__(
        self,
        field: PyField,
        source_field: PyField | None = None,
        transform: PyPartitionTransform | None = None,
    ) -> None: ...

class PyPartitionTransform:
    """Partitioning Transform from a Data Catalog source field to a Partitioning Columns."""

    @staticmethod
    def identity() -> PyPartitionTransform: ...
    @staticmethod
    def year() -> PyPartitionTransform: ...
    @staticmethod
    def month() -> PyPartitionTransform: ...
    @staticmethod
    def day() -> PyPartitionTransform: ...
    @staticmethod
    def hour() -> PyPartitionTransform: ...
    @staticmethod
    def iceberg_bucket(n: int) -> PyPartitionTransform: ...
    @staticmethod
    def iceberg_truncate(w: int) -> PyPartitionTransform: ...
    def is_identity(self) -> bool: ...
    def is_year(self) -> bool: ...
    def is_month(self) -> bool: ...
    def is_day(self) -> bool: ...
    def is_hour(self) -> bool: ...
    def is_iceberg_bucket(self) -> bool: ...
    def is_iceberg_truncate(self) -> bool: ...
    def num_buckets(self) -> int: ...
    def width(self) -> int: ...

class PyPushdowns:
    """Pushdowns from the query optimizer that can optimize scanning data sources."""

    columns: list[str] | None
    filters: PyExpr | None
    partition_filters: PyExpr | None
    limit: int | None

    def __init__(
        self,
        columns: list[str] | None = None,
        filters: PyExpr | None = None,
        partition_filters: PyExpr | None = None,
        limit: int | None = None,
    ) -> None: ...
    def filter_required_column_names(self) -> list[str]:
        """List of field names that are required by the filter predicate."""
        ...

PyArrowParquetType = tuple[pa.Field, dict[str, str], pa.Array, int]

def read_parquet(
    uri: str,
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups: list[int] | None = None,
    predicate: PyExpr | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: PyTimeUnit | None = None,
) -> PyRecordBatch: ...
def read_parquet_bulk(
    uris: list[str],
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups: list[list[int] | None] | None = None,
    predicate: PyExpr | None = None,
    io_config: IOConfig | None = None,
    num_parallel_tasks: int | None = 128,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: PyTimeUnit | None = None,
) -> list[PyRecordBatch]: ...
def read_parquet_statistics(
    uris: PySeries,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
) -> PyRecordBatch: ...
def read_parquet_into_pyarrow(
    uri: str,
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups: list[int] | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: PyTimeUnit | None = None,
    string_encoding: Literal["utf-8", "raw"] = "utf-8",
    file_timeout_ms: int | None = None,
) -> PyArrowParquetType: ...
def read_parquet_into_pyarrow_bulk(
    uris: list[str],
    columns: list[str] | None = None,
    start_offset: int | None = None,
    num_rows: int | None = None,
    row_groups: list[list[int] | None] | None = None,
    io_config: IOConfig | None = None,
    num_parallel_tasks: int | None = 128,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: PyTimeUnit | None = None,
) -> list[PyArrowParquetType]: ...
def read_parquet_schema(
    uri: str,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
    coerce_int96_timestamp_unit: PyTimeUnit | None = None,
) -> PySchema: ...
def read_csv(
    uri: str,
    convert_options: CsvConvertOptions | None = None,
    parse_options: CsvParseOptions | None = None,
    read_options: CsvReadOptions | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
) -> PyRecordBatch: ...
def read_csv_schema(
    uri: str,
    parse_options: CsvParseOptions | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
) -> PySchema: ...
def read_json(
    uri: str,
    convert_options: JsonConvertOptions | None = None,
    parse_options: JsonParseOptions | None = None,
    read_options: JsonReadOptions | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
    max_chunks_in_flight: int | None = None,
) -> PyRecordBatch: ...
def read_json_schema(
    uri: str,
    parse_options: JsonParseOptions | None = None,
    io_config: IOConfig | None = None,
    multithreaded_io: bool | None = None,
) -> PySchema: ...

class PyTimeUnit:
    @staticmethod
    def nanoseconds() -> PyTimeUnit: ...
    @staticmethod
    def microseconds() -> PyTimeUnit: ...
    @staticmethod
    def milliseconds() -> PyTimeUnit: ...
    @staticmethod
    def seconds() -> PyTimeUnit: ...
    @staticmethod
    def from_str(unit: str) -> PyTimeUnit: ...

ReduceReturnType = tuple[Callable[..., Any], tuple[Any, ...]]

class PyDataType:
    @staticmethod
    def null() -> PyDataType: ...
    @staticmethod
    def bool() -> PyDataType: ...
    @staticmethod
    def int8() -> PyDataType: ...
    @staticmethod
    def int16() -> PyDataType: ...
    @staticmethod
    def int32() -> PyDataType: ...
    @staticmethod
    def int64() -> PyDataType: ...
    @staticmethod
    def uint8() -> PyDataType: ...
    @staticmethod
    def uint16() -> PyDataType: ...
    @staticmethod
    def uint32() -> PyDataType: ...
    @staticmethod
    def uint64() -> PyDataType: ...
    @staticmethod
    def float32() -> PyDataType: ...
    @staticmethod
    def float64() -> PyDataType: ...
    @staticmethod
    def binary() -> PyDataType: ...
    @staticmethod
    def fixed_size_binary(size: int) -> PyDataType: ...
    @staticmethod
    def string() -> PyDataType: ...
    @staticmethod
    def decimal128(precision: int, size: int) -> PyDataType: ...
    @staticmethod
    def date() -> PyDataType: ...
    @staticmethod
    def time(time_unit: PyTimeUnit) -> PyDataType: ...
    @staticmethod
    def timestamp(time_unit: PyTimeUnit, timezone: str | None = None) -> PyDataType: ...
    @staticmethod
    def duration(time_unit: PyTimeUnit) -> PyDataType: ...
    @staticmethod
    def interval() -> PyDataType: ...
    @staticmethod
    def list(data_type: PyDataType) -> PyDataType: ...
    @staticmethod
    def fixed_size_list(data_type: PyDataType, size: int) -> PyDataType: ...
    @staticmethod
    def map(key_type: PyDataType, value_type: PyDataType) -> PyDataType: ...
    @staticmethod
    def struct(fields: dict[str, PyDataType]) -> PyDataType: ...
    @staticmethod
    def extension(name: str, storage_data_type: PyDataType, metadata: str | None = None) -> PyDataType: ...
    @staticmethod
    def embedding(data_type: PyDataType, size: int) -> PyDataType: ...
    @staticmethod
    def image(
        mode: ImageMode | None = None,
        height: int | None = None,
        width: int | None = None,
    ) -> PyDataType: ...
    @staticmethod
    def tensor(dtype: PyDataType, shape: tuple[int, ...] | None = None) -> PyDataType: ...
    @staticmethod
    def sparse_tensor(
        dtype: PyDataType, shape: tuple[int, ...] | None = None, use_offset_indices: builtins.bool = False
    ) -> PyDataType: ...
    @staticmethod
    def python() -> PyDataType: ...
    def to_arrow(self, cast_tensor_type_for_ray: builtins.bool | None = None) -> pa.DataType: ...
    def is_null(self) -> builtins.bool: ...
    def is_boolean(self) -> builtins.bool: ...
    def is_int8(self) -> builtins.bool: ...
    def is_int16(self) -> builtins.bool: ...
    def is_int32(self) -> builtins.bool: ...
    def is_int64(self) -> builtins.bool: ...
    def is_uint8(self) -> builtins.bool: ...
    def is_uint16(self) -> builtins.bool: ...
    def is_uint32(self) -> builtins.bool: ...
    def is_uint64(self) -> builtins.bool: ...
    def is_float32(self) -> builtins.bool: ...
    def is_float64(self) -> builtins.bool: ...
    def is_decimal128(self) -> builtins.bool: ...
    def is_timestamp(self) -> builtins.bool: ...
    def is_date(self) -> builtins.bool: ...
    def is_time(self) -> builtins.bool: ...
    def is_duration(self) -> builtins.bool: ...
    def is_interval(self) -> builtins.bool: ...
    def is_binary(self) -> builtins.bool: ...
    def is_fixed_size_binary(self) -> builtins.bool: ...
    def is_string(self) -> builtins.bool: ...
    def is_list(self) -> builtins.bool: ...
    def is_fixed_size_list(self) -> builtins.bool: ...
    def is_struct(self) -> builtins.bool: ...
    def is_map(self) -> builtins.bool: ...
    def is_extension(self) -> builtins.bool: ...
    def is_image(self) -> builtins.bool: ...
    def is_fixed_shape_image(self) -> builtins.bool: ...
    def is_embedding(self) -> builtins.bool: ...
    def is_tensor(self) -> builtins.bool: ...
    def is_fixed_shape_tensor(self) -> builtins.bool: ...
    def is_sparse_tensor(self) -> builtins.bool: ...
    def is_fixed_shape_sparse_tensor(self) -> builtins.bool: ...
    def is_python(self) -> builtins.bool: ...
    def is_numeric(self) -> builtins.bool: ...
    def is_integer(self) -> builtins.bool: ...
    def is_logical(self) -> builtins.bool: ...
    def is_temporal(self) -> builtins.bool: ...
    def fixed_size(self) -> int: ...
    def fixed_shape(self) -> tuple[int, ...]: ...
    def time_unit(self) -> PyTimeUnit: ...
    def time_zone(self) -> str | None: ...
    def dtype(self) -> PyDataType: ...
    def fields(self) -> builtins.list[PyField]: ...
    def precision(self) -> int: ...
    def scale(self) -> int: ...
    def image_mode(self) -> ImageMode | None: ...
    def use_offset_indices(self) -> builtins.bool: ...
    def key_type(self) -> PyDataType: ...
    def value_type(self) -> PyDataType: ...
    def is_equal(self, other: Any) -> builtins.bool: ...
    @staticmethod
    def from_json(serialized: str) -> PyDataType: ...
    def __reduce__(self) -> ReduceReturnType: ...
    def __hash__(self) -> int: ...

class PyField:
    def name(self) -> str: ...
    @staticmethod
    def create(name: str, datatype: PyDataType) -> PyField: ...
    def dtype(self) -> PyDataType: ...
    def eq(self, other: PyField) -> bool: ...
    def __reduce__(self) -> ReduceReturnType: ...

class PySchema:
    def __getitem__(self, name: str) -> PyField: ...
    def names(self) -> list[str]: ...
    def union(self, other: PySchema) -> PySchema: ...
    def __eq__(self, other: object) -> bool: ...
    def estimate_row_size_bytes(self) -> float: ...
    @staticmethod
    def from_field_name_and_types(names_and_types: list[tuple[str, PyDataType]]) -> PySchema: ...
    @staticmethod
    def from_fields(fields: list[PyField]) -> PySchema: ...
    def to_pyarrow_schema(self) -> pa.Schema: ...
    def __reduce__(self) -> ReduceReturnType: ...
    def __repr__(self) -> str: ...
    def _repr_html_(self) -> str: ...
    def _truncated_table_html(self) -> str: ...
    def _truncated_table_string(self) -> str: ...
    def apply_hints(self, hints: PySchema) -> PySchema: ...

class PyExpr:
    def alias(self, name: str) -> PyExpr: ...
    def cast(self, dtype: PyDataType) -> PyExpr: ...
    def if_else(self, if_true: PyExpr, if_false: PyExpr) -> PyExpr: ...
    def count(self, mode: CountMode) -> PyExpr: ...
    def count_distinct(self) -> PyExpr: ...
    def sum(self) -> PyExpr: ...
    def approx_count_distinct(self) -> PyExpr: ...
    def approx_percentiles(self, percentiles: float | list[float]) -> PyExpr: ...
    def mean(self) -> PyExpr: ...
    def stddev(self) -> PyExpr: ...
    def min(self) -> PyExpr: ...
    def max(self) -> PyExpr: ...
    def bool_and(self) -> PyExpr: ...
    def bool_or(self) -> PyExpr: ...
    def any_value(self, ignore_nulls: bool) -> PyExpr: ...
    def skew(self) -> PyExpr: ...
    def agg_list(self) -> PyExpr: ...
    def agg_set(self) -> PyExpr: ...
    def agg_concat(self) -> PyExpr: ...
    def over(self, window_spec: WindowSpec) -> PyExpr: ...
    def offset(self, offset: int, default: PyExpr | None = None) -> PyExpr: ...
    def __add__(self, other: PyExpr) -> PyExpr: ...
    def __sub__(self, other: PyExpr) -> PyExpr: ...
    def __mul__(self, other: PyExpr) -> PyExpr: ...
    def __floordiv__(self, other: PyExpr) -> PyExpr: ...
    def __truediv__(self, other: PyExpr) -> PyExpr: ...
    def __mod__(self, other: PyExpr) -> PyExpr: ...
    def __and__(self, other: PyExpr) -> PyExpr: ...
    def __or__(self, other: PyExpr) -> PyExpr: ...
    def __xor__(self, other: PyExpr) -> PyExpr: ...
    def __invert__(self) -> PyExpr: ...
    def __lt__(self, other: PyExpr) -> PyExpr: ...
    def __le__(self, other: PyExpr) -> PyExpr: ...
    def __gt__(self, other: PyExpr) -> PyExpr: ...
    def __ge__(self, other: PyExpr) -> PyExpr: ...
    def __eq__(self, other: PyExpr) -> PyExpr: ...  # type: ignore[override]
    def __ne__(self, other: PyExpr) -> PyExpr: ...  # type: ignore[override]
    def __rshift__(self, other: PyExpr) -> PyExpr: ...
    def __lshift__(self, other: PyExpr) -> PyExpr: ...
    def is_null(self) -> PyExpr: ...
    def not_null(self) -> PyExpr: ...
    def fill_null(self, fill_value: PyExpr) -> PyExpr: ...
    def eq_null_safe(self, other: PyExpr) -> PyExpr: ...
    def is_in(self, other: list[PyExpr]) -> PyExpr: ...
    def between(self, lower: PyExpr, upper: PyExpr) -> PyExpr: ...
    def name(self) -> str: ...
    def to_field(self, schema: PySchema) -> PyField: ...
    def to_sql(self) -> str: ...
    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __reduce__(self) -> ReduceReturnType: ...
    def struct_get(self, name: str) -> PyExpr: ...
    def map_get(self, key: PyExpr) -> PyExpr: ...
    def partitioning_days(self) -> PyExpr: ...
    def partitioning_hours(self) -> PyExpr: ...
    def partitioning_months(self) -> PyExpr: ...
    def partitioning_years(self) -> PyExpr: ...
    def partitioning_iceberg_bucket(self, n: int) -> PyExpr: ...
    def partitioning_iceberg_truncate(self, w: int) -> PyExpr: ...

    ###
    # Visitor methods
    ###

    def accept(self, visitor: ExpressionVisitor[R]) -> R: ...

    ###
    # Helper methods from Expr from Eq Hash traits
    ###

    def _eq(self, other: PyExpr) -> bool: ...
    def _ne(self, other: PyExpr) -> bool: ...
    def _hash(self) -> int: ...

    ###
    # Helper methods required by optimizer:
    # These should be removed from the Python API for Expressions when logical plans and optimizer are migrated to Rust
    ###
    def _input_mapping(self) -> builtins.str | None: ...

def eq(expr1: PyExpr, expr2: PyExpr) -> bool: ...
def unresolved_col(name: str) -> PyExpr: ...
def resolved_col(name: str) -> PyExpr: ...
def lit(item: Any) -> PyExpr: ...
def list_(items: list[PyExpr]) -> PyExpr: ...
def date_lit(item: int) -> PyExpr: ...
def time_lit(item: int, tu: PyTimeUnit) -> PyExpr: ...
def timestamp_lit(item: int, tu: PyTimeUnit, tz: str | None) -> PyExpr: ...
def duration_lit(item: int, tu: PyTimeUnit) -> PyExpr: ...
def interval_lit(
    years: int | None,
    months: int | None,
    days: int | None,
    hours: int | None,
    minutes: int | None,
    seconds: int | None,
    millis: int | None,
    nanos: int | None,
) -> PyExpr: ...
def decimal_lit(sign: bool, digits: tuple[int, ...], exp: int) -> PyExpr: ...
def series_lit(item: PySeries) -> PyExpr: ...
def udf(
    name: str,
    inner: UninitializedUdf,
    bound_args: BoundUDFArgs,
    expressions: list[PyExpr],
    return_dtype: PyDataType,
    init_args: InitArgsType,
    resource_request: ResourceRequest | None,
    batch_size: int | None,
    concurrency: int | None,
) -> PyExpr: ...
def initialize_udfs(expression: PyExpr) -> PyExpr: ...
def get_udf_names(expression: PyExpr) -> list[str]: ...
def resolve_expr(expr: PyExpr, schema: PySchema) -> tuple[PyExpr, PyField]: ...
def row_number() -> PyExpr: ...
def rank() -> PyExpr: ...
def dense_rank() -> PyExpr: ...

# -----
# SQL functions
# -----
class SQLFunctionStub:
    @property
    def name(self) -> str: ...
    @property
    def docstring(self) -> str: ...
    @property
    def arg_names(self) -> list[str]: ...

def sql_exec(
    source: str, session: PySession, ctes: dict[str, LogicalPlanBuilder], config: PyDaftPlanningConfig
) -> object | None: ...
def sql_expr(sql: str) -> PyExpr: ...
def sql_datatype(sql: str) -> PyDataType: ...
def list_sql_functions() -> list[SQLFunctionStub]: ...
def connect_start(addr: str = "sc://0.0.0.0:0") -> ConnectionHandle: ...

class ConnectionHandle:
    def shutdown(self) -> None: ...
    def port(self) -> int: ...

# ---
# expr.list namespace
# ---
def list_sort(expr: PyExpr, desc: PyExpr, nulls_first: PyExpr) -> PyExpr: ...
def list_distinct(expr: PyExpr) -> PyExpr: ...
def list_value_counts(expr: PyExpr) -> PyExpr: ...
def list_join(expr: PyExpr, delimiter: PyExpr) -> PyExpr: ...
def list_count(expr: PyExpr, mode: CountMode) -> PyExpr: ...
def list_get(expr: PyExpr, idx: PyExpr, default: PyExpr) -> PyExpr: ...
def list_sum(expr: PyExpr) -> PyExpr: ...
def list_mean(expr: PyExpr) -> PyExpr: ...
def list_min(expr: PyExpr) -> PyExpr: ...
def list_max(expr: PyExpr) -> PyExpr: ...
def list_bool_and(expr: PyExpr) -> PyExpr: ...
def list_bool_or(expr: PyExpr) -> PyExpr: ...
def list_slice(expr: PyExpr, start: PyExpr, end: PyExpr | None = None) -> PyExpr: ...
def list_chunk(expr: PyExpr, size: int) -> PyExpr: ...

class PySqlCatalog:
    @staticmethod
    def new() -> PySqlCatalog: ...
    def register_table(self, name: str, logical_plan_builder: LogicalPlanBuilder) -> None: ...
    def copy_from(self, other: PySqlCatalog) -> None: ...
    def to_pydict(self) -> dict[str, LogicalPlanBuilder]: ...

class PySeries:
    @staticmethod
    def from_arrow(name: str, pyarrow_array: pa.Array) -> PySeries: ...
    @staticmethod
    def from_pylist(name: str, pylist: list[Any], pyobj: str) -> PySeries: ...
    def to_pylist(self) -> list[Any]: ...
    def to_arrow(self) -> pa.Array: ...
    def __abs__(self) -> PySeries: ...
    def __add__(self, other: PySeries) -> PySeries: ...
    def __sub__(self, other: PySeries) -> PySeries: ...
    def __mul__(self, other: PySeries) -> PySeries: ...
    def __truediv__(self, other: PySeries) -> PySeries: ...
    def __mod__(self, other: PySeries) -> PySeries: ...
    def __and__(self, other: PySeries) -> PySeries: ...
    def __or__(self, other: PySeries) -> PySeries: ...
    def __xor__(self, other: PySeries) -> PySeries: ...
    def __lt__(self, other: PySeries) -> PySeries: ...
    def __le__(self, other: PySeries) -> PySeries: ...
    def __gt__(self, other: PySeries) -> PySeries: ...
    def __ge__(self, other: PySeries) -> PySeries: ...
    def __eq__(self, other: PySeries) -> PySeries: ...  # type: ignore[override]
    def __ne__(self, other: PySeries) -> PySeries: ...  # type: ignore[override]
    def __rshift__(self, other: PySeries) -> PySeries: ...
    def __lshift__(self, other: PySeries) -> PySeries: ...
    def __floordiv__(self, other: PySeries) -> PySeries: ...
    def take(self, idx: PySeries) -> PySeries: ...
    def slice(self, start: int, end: int) -> PySeries: ...
    def filter(self, mask: PySeries) -> PySeries: ...
    def sort(self, descending: bool, nulls_first: bool) -> PySeries: ...
    def argsort(self, descending: bool, nulls_first: bool) -> PySeries: ...
    def hash(self, seed: PySeries | None = None) -> PySeries: ...
    def minhash(
        self,
        num_hashes: int,
        ngram_size: int,
        seed: int = 1,
        hash_function: Literal["murmurhash3", "xxhash", "sha1"] = "murmurhash3",
    ) -> PySeries: ...
    def __invert__(self) -> PySeries: ...
    def count(self, mode: CountMode) -> PySeries: ...
    def count_distinct(self) -> PySeries: ...
    def sum(self) -> PySeries: ...
    def mean(self) -> PySeries: ...
    def stddev(self) -> PySeries: ...
    def min(self) -> PySeries: ...
    def max(self) -> PySeries: ...
    def agg_list(self) -> PySeries: ...
    def agg_set(self) -> PySeries: ...
    def cast(self, dtype: PyDataType) -> PySeries: ...
    def log2(self) -> PySeries: ...
    def log10(self) -> PySeries: ...
    def log(self, base: float) -> PySeries: ...
    def ln(self) -> PySeries: ...
    def log1p(self) -> PySeries: ...
    @staticmethod
    def concat(series: list[PySeries]) -> PySeries: ...
    def __len__(self) -> int: ...
    def size_bytes(self) -> int: ...
    def name(self) -> str: ...
    def rename(self, name: str) -> PySeries: ...
    def data_type(self) -> PyDataType: ...
    def partitioning_days(self) -> PySeries: ...
    def partitioning_hours(self) -> PySeries: ...
    def partitioning_months(self) -> PySeries: ...
    def partitioning_years(self) -> PySeries: ...
    def partitioning_iceberg_bucket(self, n: int) -> PySeries: ...
    def partitioning_iceberg_truncate(self, w: int) -> PySeries: ...
    def list_count(self, mode: CountMode) -> PySeries: ...
    def list_get(self, idx: PySeries, default: PySeries) -> PySeries: ...
    def list_slice(self, start: PySeries, end: PySeries | None = None) -> PySeries: ...
    def list_sort(self, desc: PySeries, nulls_first: PySeries) -> PySeries: ...
    def map_get(self, key: PySeries) -> PySeries: ...
    def if_else(self, other: PySeries, predicate: PySeries) -> PySeries: ...
    def is_null(self) -> PySeries: ...
    def not_null(self) -> PySeries: ...
    def fill_null(self, fill_value: PySeries) -> PySeries: ...
    def murmur3_32(self) -> PySeries: ...
    def to_str_values(self) -> PySeries: ...
    def _debug_bincode_serialize(self) -> bytes: ...
    @staticmethod
    def _debug_bincode_deserialize(b: bytes) -> PySeries: ...

class PyShowOptions:
    pass

class PyRecordBatch:
    def schema(self) -> PySchema: ...
    def eval_expression_list(self, exprs: list[PyExpr]) -> PyRecordBatch: ...
    def take(self, idx: PySeries) -> PyRecordBatch: ...
    def filter(self, exprs: list[PyExpr]) -> PyRecordBatch: ...
    def sort(self, sort_keys: list[PyExpr], descending: list[bool], nulls_first: list[bool]) -> PyRecordBatch: ...
    def argsort(self, sort_keys: list[PyExpr], descending: list[bool], nulls_first: list[bool]) -> PySeries: ...
    def agg(self, to_agg: list[PyExpr], group_by: list[PyExpr]) -> PyRecordBatch: ...
    def pivot(
        self,
        group_by: list[PyExpr],
        pivot_column: PyExpr,
        values_column: PyExpr,
        names: list[str],
    ) -> PyRecordBatch: ...
    def hash_join(
        self,
        right: PyRecordBatch,
        left_on: list[PyExpr],
        right_on: list[PyExpr],
        how: JoinType,
    ) -> PyRecordBatch: ...
    def sort_merge_join(
        self,
        right: PyRecordBatch,
        left_on: list[PyExpr],
        right_on: list[PyExpr],
        is_sorted: bool,
    ) -> PyRecordBatch: ...
    def explode(self, to_explode: list[PyExpr]) -> PyRecordBatch: ...
    def head(self, num: int) -> PyRecordBatch: ...
    def sample_by_fraction(self, fraction: float, with_replacement: bool, seed: int | None) -> PyRecordBatch: ...
    def sample_by_size(self, size: int, with_replacement: bool, seed: int | None) -> PyRecordBatch: ...
    def quantiles(self, num: int) -> PyRecordBatch: ...
    def partition_by_hash(self, exprs: list[PyExpr], num_partitions: int) -> list[PyRecordBatch]: ...
    def partition_by_random(self, num_partitions: int, seed: int) -> list[PyRecordBatch]: ...
    def partition_by_range(
        self, partition_keys: list[PyExpr], boundaries: PyRecordBatch, descending: list[bool]
    ) -> list[PyRecordBatch]: ...
    def partition_by_value(self, partition_keys: list[PyExpr]) -> tuple[list[PyRecordBatch], PyRecordBatch]: ...
    def add_monotonically_increasing_id(self, partition_num: int, column_name: str) -> PyRecordBatch: ...
    def preview(self, format: str | None, options: str | None) -> str: ...
    def __repr__(self) -> str: ...
    def _repr_html_(self) -> str: ...
    def __len__(self) -> int: ...
    def size_bytes(self) -> int: ...
    def get_column(self, idx: int) -> PySeries: ...
    def columns(self) -> list[PySeries]: ...
    @staticmethod
    def concat(tables: list[PyRecordBatch]) -> PyRecordBatch: ...
    def slice(self, start: int, end: int) -> PyRecordBatch: ...
    @staticmethod
    def from_arrow_record_batches(record_batches: list[pa.RecordBatch], schema: PySchema) -> PyRecordBatch: ...
    @staticmethod
    def from_pylist_series(dict: dict[str, PySeries]) -> PyRecordBatch: ...
    @staticmethod
    def from_pyseries_list(list: list[PySeries]) -> PyRecordBatch: ...
    def to_arrow_record_batch(self) -> pa.RecordBatch: ...
    @staticmethod
    def empty(schema: PySchema | None = None) -> PyRecordBatch: ...
    @staticmethod
    def from_file_infos(file_infos: FileInfos) -> PyRecordBatch: ...
    def to_file_infos(self) -> FileInfos: ...

class PyMicroPartition:
    def schema(self) -> PySchema: ...
    def column_names(self) -> list[str]: ...
    def get_column_by_name(self, name: str) -> PySeries: ...
    def get_column(self, idx: int) -> PySeries: ...
    def columns(self) -> list[PySeries]: ...
    def get_record_batches(self) -> list[PyRecordBatch]: ...
    def size_bytes(self) -> int | None: ...
    def _repr_html_(self) -> str: ...
    @staticmethod
    def empty(schema: PySchema | None = None) -> PyMicroPartition: ...
    @staticmethod
    def from_scan_task(scan_task: ScanTask) -> PyMicroPartition: ...
    @staticmethod
    def from_record_batches(record_batches: list[PyRecordBatch]) -> PyMicroPartition: ...
    @staticmethod
    def from_arrow_record_batches(record_batches: list[pa.RecordBatch], schema: PySchema) -> PyMicroPartition: ...
    @staticmethod
    def concat(tables: list[PyMicroPartition]) -> PyMicroPartition: ...
    @staticmethod
    def read_from_ipc_stream(bytes: bytes) -> PyMicroPartition: ...
    def write_to_ipc_stream(self) -> bytes: ...
    def slice(self, start: int, end: int) -> PyMicroPartition: ...
    def to_record_batch(self) -> PyRecordBatch: ...
    def cast_to_schema(self, schema: PySchema) -> PyMicroPartition: ...
    def eval_expression_list(self, exprs: list[PyExpr]) -> PyMicroPartition: ...
    def take(self, idx: PySeries) -> PyMicroPartition: ...
    def filter(self, exprs: list[PyExpr]) -> PyMicroPartition: ...
    def sort(self, sort_keys: list[PyExpr], descending: list[bool], nulls_first: list[bool]) -> PyMicroPartition: ...
    def argsort(self, sort_keys: list[PyExpr], descending: list[bool], nulls_first: list[bool]) -> PySeries: ...
    def agg(self, to_agg: list[PyExpr], group_by: list[PyExpr]) -> PyMicroPartition: ...
    def hash_join(
        self,
        right: PyMicroPartition,
        left_on: list[PyExpr],
        right_on: list[PyExpr],
        how: JoinType,
        null_equals_nulls: list[bool] | None = None,
    ) -> PyMicroPartition: ...
    def pivot(
        self,
        group_by: list[PyExpr],
        pivot_column: PyExpr,
        values_column: PyExpr,
        names: list[str],
    ) -> PyMicroPartition: ...
    def sort_merge_join(
        self,
        right: PyMicroPartition,
        left_on: list[PyExpr],
        right_on: list[PyExpr],
        is_sorted: bool,
    ) -> PyMicroPartition: ...
    def cross_join(
        self,
        right: PyMicroPartition,
        outer_loop_side: JoinSide,
    ) -> PyMicroPartition: ...
    def explode(self, to_explode: list[PyExpr]) -> PyMicroPartition: ...
    def unpivot(
        self,
        ids: list[PyExpr],
        values: list[PyExpr],
        variable_name: str,
        value_name: str,
    ) -> PyMicroPartition: ...
    def head(self, num: int) -> PyMicroPartition: ...
    def sample_by_fraction(self, fraction: float, with_replacement: bool, seed: int | None) -> PyMicroPartition: ...
    def sample_by_size(self, size: int, with_replacement: bool, seed: int | None) -> PyMicroPartition: ...
    def quantiles(self, num: int) -> PyMicroPartition: ...
    def partition_by_hash(self, exprs: list[PyExpr], num_partitions: int) -> list[PyMicroPartition]: ...
    def partition_by_random(self, num_partitions: int, seed: int) -> list[PyMicroPartition]: ...
    def partition_by_range(
        self, partition_keys: list[PyExpr], boundaries: PyRecordBatch, descending: list[bool]
    ) -> list[PyMicroPartition]: ...
    def partition_by_value(self, exprs: list[PyExpr]) -> tuple[list[PyMicroPartition], PyMicroPartition]: ...
    def add_monotonically_increasing_id(self, partition_num: int, column_name: str) -> PyMicroPartition: ...
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    @classmethod
    def read_parquet(
        cls,
        path: str,
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups: list[int] | None = None,
        predicate: PyExpr | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: PyTimeUnit = PyTimeUnit.nanoseconds(),
    ) -> PyMicroPartition: ...
    @classmethod
    def read_parquet_bulk(
        cls,
        uris: list[str],
        columns: list[str] | None = None,
        start_offset: int | None = None,
        num_rows: int | None = None,
        row_groups: list[list[int] | None] | None = None,
        predicate: PyExpr | None = None,
        io_config: IOConfig | None = None,
        num_parallel_tasks: int | None = None,
        multithreaded_io: bool | None = None,
        coerce_int96_timestamp_unit: PyTimeUnit | None = None,
    ) -> PyMicroPartition: ...
    @classmethod
    def read_csv(
        cls,
        uri: str,
        convert_options: CsvConvertOptions | None = None,
        parse_options: CsvParseOptions | None = None,
        read_options: CsvReadOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> PyMicroPartition: ...
    @classmethod
    def read_json_native(
        cls,
        uri: str,
        convert_options: JsonConvertOptions | None = None,
        parse_options: JsonParseOptions | None = None,
        read_options: JsonReadOptions | None = None,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> PyMicroPartition: ...
    @classmethod
    def read_warc(
        cls,
        uri: str,
        io_config: IOConfig | None = None,
        multithreaded_io: bool | None = None,
    ) -> PyMicroPartition: ...

class PyMicroPartitionSet:
    def get_partition(self, idx: int) -> PyMicroPartition: ...
    def set_partition(self, idx: int, part: PyMicroPartition) -> None: ...
    def delete_partition(self, idx: int) -> None: ...
    def has_partition(self, idx: int) -> bool: ...
    def __len__(self) -> int: ...
    def size_bytes(self) -> int | None: ...
    def num_partitions(self) -> int: ...
    def wait(self) -> None: ...

class PhysicalPlanScheduler:
    """A work scheduler for physical query plans."""

    @staticmethod
    def from_logical_plan_builder(
        logical_plan_builder: LogicalPlanBuilder,
        cfg: PyDaftExecutionConfig,
    ) -> PhysicalPlanScheduler: ...
    def num_partitions(self) -> int: ...
    def repr_ascii(self, simple: bool) -> str: ...
    def repr_mermaid(self, options: MermaidOptions) -> str: ...
    def to_json_string(self) -> str: ...
    def to_partition_tasks(
        self, psets: dict[str, list[PartitionT]], actor_pool_manager: Any
    ) -> physical_plan.InProgressPhysicalPlan[PartitionT]: ...
    def run(self, psets: dict[str, list[PartitionT]]) -> Iterator[PyMicroPartition]: ...

class AdaptivePhysicalPlanScheduler:
    """An adaptive Physical Plan Scheduler."""

    @staticmethod
    def from_logical_plan_builder(
        logical_plan_builder: LogicalPlanBuilder,
        cfg: PyDaftExecutionConfig,
    ) -> AdaptivePhysicalPlanScheduler: ...
    def next(self) -> tuple[int | None, PhysicalPlanScheduler]: ...
    def is_done(self) -> bool: ...
    # Todo use in memory info here instead
    def update(
        self,
        stage_id: int,
        partition_key: str,
        cache_entry: PartitionCacheEntry,
        num_partitions: int,
        size_bytes: int,
        num_rows: int,
    ) -> None: ...
    def update_stats(
        self, time_taken: float, size_bytes: int | None, num_rows: int | None, stage_id: int | None
    ) -> None: ...
    def explain_analyze(self, explain_analyze_dir: str) -> None: ...

class LogicalPlanBuilder:
    """A logical plan builder, which simplifies constructing logical plans via a fluent interface.

    E.g., LogicalPlanBuilder.table_scan(..).project(..).filter(..).

    This builder holds the current root (sink) of the logical plan, and the building methods return
    a brand new builder holding a new plan; i.e., this is an immutable builder.
    """

    @staticmethod
    def in_memory_scan(
        partition_key: str,
        cache_entry: PartitionCacheEntry,
        schema: PySchema,
        num_partitions: int,
        size_bytes: int,
        num_rows: int,
    ) -> LogicalPlanBuilder: ...
    def with_planning_config(self, daft_planning_config: PyDaftPlanningConfig) -> LogicalPlanBuilder: ...
    def select(self, to_select: list[PyExpr]) -> LogicalPlanBuilder: ...
    def with_columns(self, columns: list[PyExpr]) -> LogicalPlanBuilder: ...
    def with_columns_renamed(self, cols_map: dict[str, str]) -> LogicalPlanBuilder: ...
    def exclude(self, to_exclude: list[str]) -> LogicalPlanBuilder: ...
    def filter(self, predicate: PyExpr) -> LogicalPlanBuilder: ...
    def limit(self, limit: int, eager: bool) -> LogicalPlanBuilder: ...
    def explode(self, to_explode: list[PyExpr]) -> LogicalPlanBuilder: ...
    def unpivot(
        self,
        ids: list[PyExpr],
        values: list[PyExpr],
        variable_name: str,
        value_name: str,
    ) -> LogicalPlanBuilder: ...
    def sort(self, sort_by: list[PyExpr], descending: list[bool], nulls_first: list[bool]) -> LogicalPlanBuilder: ...
    def hash_repartition(
        self,
        partition_by: list[PyExpr],
        num_partitions: int | None,
    ) -> LogicalPlanBuilder: ...
    def random_shuffle(self, num_partitions: int | None) -> LogicalPlanBuilder: ...
    def into_partitions(self, num_partitions: int) -> LogicalPlanBuilder: ...
    def coalesce(self, num_partitions: int) -> LogicalPlanBuilder: ...
    def distinct(self) -> LogicalPlanBuilder: ...
    def sample(self, fraction: float, with_replacement: bool, seed: int | None) -> LogicalPlanBuilder: ...
    def aggregate(self, agg_exprs: list[PyExpr], groupby_exprs: list[PyExpr]) -> LogicalPlanBuilder: ...
    def pivot(
        self,
        groupby_exprs: list[PyExpr],
        pivot_expr: PyExpr,
        values_expr: PyExpr,
        agg_expr: PyExpr,
        names: list[str],
    ) -> LogicalPlanBuilder: ...
    def join(
        self,
        right: LogicalPlanBuilder,
        left_on: list[PyExpr],
        right_on: list[PyExpr],
        join_type: JoinType,
        join_strategy: JoinStrategy | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
    ) -> LogicalPlanBuilder: ...
    def concat(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder: ...
    def union(self, other: LogicalPlanBuilder, is_all: bool, is_by_name: bool) -> LogicalPlanBuilder: ...
    def intersect(self, other: LogicalPlanBuilder, is_all: bool) -> LogicalPlanBuilder: ...
    def except_(self, other: LogicalPlanBuilder, is_all: bool) -> LogicalPlanBuilder: ...
    def add_monotonically_increasing_id(self, column_name: str | None) -> LogicalPlanBuilder: ...
    def table_write(
        self,
        root_dir: str,
        write_mode: WriteMode,
        file_format: FileFormat,
        partition_cols: list[PyExpr] | None = None,
        compression: str | None = None,
        io_config: IOConfig | None = None,
    ) -> LogicalPlanBuilder: ...
    def iceberg_write(
        self,
        table_name: str,
        table_location: str,
        partition_spec_id: int,
        partition_cols: list[PyExpr],
        iceberg_schema: IcebergSchema,
        iceberg_properties: IcebergTableProperties,
        catalog_columns: list[str],
        io_config: IOConfig | None = None,
    ) -> LogicalPlanBuilder: ...
    def delta_write(
        self,
        path: str,
        columns_name: list[str],
        mode: str,
        version: int,
        large_dtypes: bool,
        partition_cols: list[str] | None = None,
        io_config: IOConfig | None = None,
    ) -> LogicalPlanBuilder: ...
    def lance_write(
        self,
        path: str,
        columns_name: list[str],
        mode: str,
        io_config: IOConfig | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> LogicalPlanBuilder: ...
    def datasink_write(self, name: str, sink: DataSink[WriteResultType]) -> LogicalPlanBuilder: ...
    def schema(self) -> PySchema: ...
    def describe(self) -> LogicalPlanBuilder: ...
    def summarize(self) -> LogicalPlanBuilder: ...
    def optimize(self) -> LogicalPlanBuilder: ...
    def to_physical_plan_scheduler(self, cfg: PyDaftExecutionConfig) -> PhysicalPlanScheduler: ...
    def to_adaptive_physical_plan_scheduler(self, cfg: PyDaftExecutionConfig) -> AdaptivePhysicalPlanScheduler: ...
    def repr_ascii(self, simple: bool) -> str: ...
    def repr_mermaid(self, options: MermaidOptions) -> str: ...
    def repr_json(self, include_schema: bool) -> str: ...

class DistributedPhysicalPlan:
    @staticmethod
    def from_logical_plan_builder(
        builder: LogicalPlanBuilder, config: PyDaftExecutionConfig
    ) -> DistributedPhysicalPlan: ...
    def id(self) -> str: ...
    def repr_ascii(self, simple: bool) -> str: ...
    def repr_mermaid(self, options: MermaidOptions) -> str: ...

class DistributedPhysicalPlanRunner:
    def __init__(self) -> None: ...
    def run_plan(
        self, plan: DistributedPhysicalPlan, psets: dict[str, list[RayPartitionRef]]
    ) -> AsyncIterator[tuple[object, int, int]]: ...

class LocalPhysicalPlan:
    @staticmethod
    def from_logical_plan_builder(builder: LogicalPlanBuilder) -> LocalPhysicalPlan: ...

class RayPartitionRef:
    object_ref: ray.ObjectRef
    num_rows: int
    size_bytes: int

    def __init__(self, object_ref: ray.ObjectRef, num_rows: int, size_bytes: int): ...

class RaySwordfishTask:
    def name(self) -> str: ...
    def plan(self) -> LocalPhysicalPlan: ...
    def psets(self) -> dict[str, list[RayPartitionRef]]: ...
    def config(self) -> PyDaftExecutionConfig: ...
    def context(self) -> dict[str, str]: ...

class RaySwordfishWorker:
    def __init__(
        self,
        worker_id: str,
        actor_handle: RaySwordfishActorHandle,
        num_cpus: int,
        num_gpus: int,
        total_memory_bytes: int,
    ) -> None: ...

class NativeExecutor:
    def __init__(self) -> None: ...
    def run(
        self,
        plan: LocalPhysicalPlan,
        psets: dict[str, list[PyMicroPartition]],
        daft_execution_config: PyDaftExecutionConfig,
        results_buffer_size: int | None,
    ) -> Iterator[PyMicroPartition]: ...
    def run_async(
        self,
        plan: LocalPhysicalPlan,
        psets: dict[str, list[PyMicroPartition]],
        daft_execution_config: PyDaftExecutionConfig,
        results_buffer_size: int | None,
        context: dict[str, str] | None,
    ) -> AsyncIterator[PyMicroPartition]: ...
    def repr_ascii(
        self, builder: LogicalPlanBuilder, daft_execution_config: PyDaftExecutionConfig, simple: bool
    ) -> str: ...
    def repr_mermaid(
        self, builder: LogicalPlanBuilder, daft_execution_config: PyDaftExecutionConfig, options: MermaidOptions
    ) -> str: ...
    def get_relationship_info(
        self,
        logical_plan_builder: LogicalPlanBuilder,
        daft_execution_config: PyDaftExecutionConfig,
    ) -> RelationshipInformation: ...

class RelationshipInformation:
    ids: list[RelationshipNode]
    plan_id: str

class RelationshipNode:
    id: int
    parent_id: int | None

class PyDaftExecutionConfig:
    @staticmethod
    def from_env() -> PyDaftExecutionConfig: ...
    def with_config_values(
        self,
        scan_tasks_min_size_bytes: int | None = None,
        scan_tasks_max_size_bytes: int | None = None,
        max_sources_per_scan_task: int | None = None,
        broadcast_join_size_bytes_threshold: int | None = None,
        parquet_split_row_groups_max_files: int | None = None,
        sort_merge_join_sort_with_aligned_boundaries: bool | None = None,
        hash_join_partition_size_leniency: float | None = None,
        sample_size_for_sort: int | None = None,
        num_preview_rows: int | None = None,
        parquet_target_filesize: int | None = None,
        parquet_target_row_group_size: int | None = None,
        parquet_inflation_factor: float | None = None,
        csv_target_filesize: int | None = None,
        csv_inflation_factor: float | None = None,
        shuffle_aggregation_default_partitions: int | None = None,
        partial_aggregation_threshold: int | None = None,
        high_cardinality_aggregation_threshold: float | None = None,
        read_sql_partition_size_bytes: int | None = None,
        enable_aqe: bool | None = None,
        default_morsel_size: int | None = None,
        enable_ray_tracing: bool | None = None,
        shuffle_algorithm: str | None = None,
        pre_shuffle_merge_threshold: int | None = None,
        flight_shuffle_dirs: list[str] | None = None,
        scantask_splitting_level: int | None = None,
        native_parquet_writer: bool | None = None,
        flotilla: bool | None = None,
        min_cpu_per_task: float | None = None,
    ) -> PyDaftExecutionConfig: ...
    @property
    def scan_tasks_min_size_bytes(self) -> int: ...
    @property
    def scan_tasks_max_size_bytes(self) -> int: ...
    @property
    def max_sources_per_scan_task(self) -> int: ...
    @property
    def broadcast_join_size_bytes_threshold(self) -> int: ...
    @property
    def sort_merge_join_sort_with_aligned_boundaries(self) -> bool: ...
    @property
    def hash_join_partition_size_leniency(self) -> float: ...
    @property
    def sample_size_for_sort(self) -> int: ...
    @property
    def num_preview_rows(self) -> int: ...
    @property
    def parquet_target_filesize(self) -> int: ...
    @property
    def parquet_target_row_group_size(self) -> int: ...
    @property
    def parquet_inflation_factor(self) -> float: ...
    @property
    def csv_target_filesize(self) -> int: ...
    @property
    def csv_inflation_factor(self) -> float: ...
    @property
    def shuffle_aggregation_default_partitions(self) -> int: ...
    @property
    def partial_aggregation_threshold(self) -> int: ...
    @property
    def high_cardinality_aggregation_threshold(self) -> float: ...
    @property
    def read_sql_partition_size_bytes(self) -> int: ...
    @property
    def enable_aqe(self) -> bool: ...
    @property
    def default_morsel_size(self) -> int: ...
    @property
    def shuffle_algorithm(self) -> str: ...
    @property
    def pre_shuffle_merge_threshold(self) -> int: ...
    @property
    def flight_shuffle_dirs(self) -> list[str]: ...
    @property
    def enable_ray_tracing(self) -> bool: ...
    @property
    def flotilla(self) -> bool: ...
    @property
    def min_cpu_per_task(self) -> float: ...

class PyDaftPlanningConfig:
    @staticmethod
    def from_env() -> PyDaftPlanningConfig: ...
    def with_config_values(
        self,
        default_io_config: IOConfig | None = None,
    ) -> PyDaftPlanningConfig: ...
    @property
    def default_io_config(self) -> IOConfig: ...

class PyDaftContext:
    def __init__(self) -> None: ...
    _runner: Runner[Any]
    def get_or_create_runner(self) -> Runner[PartitionT]: ...
    _daft_execution_config: PyDaftExecutionConfig
    _daft_planning_config: PyDaftPlanningConfig
    @property
    def daft_execution_config(self) -> PyDaftExecutionConfig: ...
    @property
    def daft_planning_config(self) -> PyDaftPlanningConfig: ...

def set_runner_ray(
    address: str | None = None,
    noop_if_initialized: bool = False,
    max_task_backlog: int | None = None,
    force_client_mode: bool = False,
) -> PyDaftContext: ...
def set_runner_native(num_threads: int | None = None) -> PyDaftContext: ...
def get_context() -> PyDaftContext: ...
def build_type() -> str: ...
def version() -> str: ...
def refresh_logger() -> None: ...
def get_max_log_level() -> str: ...
def set_compute_runtime_num_worker_threads(num_threads: int) -> None: ...
def io_glob(
    path: str,
    multithreaded_io: bool | None = None,
    io_config: IOConfig | None = None,
    fanout_limit: int | None = None,
    page_size: int | None = None,
    limit: int | None = None,
) -> list[dict[str, Any]]: ...

class SystemInfo:
    """Accessor for system information."""

    def __init__(self) -> None: ...
    def total_memory(self) -> int: ...
    def cpu_count(self) -> int | None: ...

###
# daft-catalog
###

class PyCatalog:
    def name(self) -> str: ...
    def create_namespace(self, ident: PyIdentifier) -> None: ...
    def create_table(self, ident: PyIdentifier, schema: PySchema) -> Table: ...
    def drop_namespace(self, ident: PyIdentifier) -> None: ...
    def drop_table(self, ident: PyIdentifier) -> None: ...
    def get_table(self, ident: PyIdentifier) -> Table: ...
    def has_table(self, ident: PyIdentifier) -> bool: ...
    def has_namespace(self, ident: PyIdentifier) -> bool: ...
    def list_namespaces(self, pattern: str | None = None) -> list[PyIdentifier]: ...
    def list_tables(self, pattern: str | None = None) -> list[PyIdentifier]: ...
    @staticmethod
    def new_memory_catalog(name: str) -> Catalog: ...

class PyTable:
    def name(self) -> str: ...
    def schema(self) -> PySchema: ...
    def to_logical_plan(self) -> LogicalPlanBuilder: ...
    def append(self, plan: LogicalPlanBuilder, **options: Any) -> None: ...
    def overwrite(self, plan: LogicalPlanBuilder, **options: Any) -> None: ...
    @staticmethod
    def new_memory_table(name: str, schema: PySchema) -> Table: ...

class PyIdentifier:
    def __init__(self, parts: tuple[str, ...]): ...
    @staticmethod
    def from_sql(input: str, normalize: bool) -> PyIdentifier: ...
    def eq(self, other: PyIdentifier) -> bool: ...
    def getitem(self, index: int) -> str: ...
    def __len__(self) -> int: ...
    def __repr__(self) -> str: ...

class PyTableSource:
    @staticmethod
    def from_pyschema(schema: PySchema) -> PyTableSource: ...
    @staticmethod
    def from_pybuilder(builder: LogicalPlanBuilder) -> PyTableSource: ...

###
# daft-session
###

class PySession:
    def __init__(self) -> None: ...
    @staticmethod
    def empty() -> PySession: ...
    def attach_catalog(self, catalog: Catalog, alias: str) -> None: ...
    def attach_table(self, table: Table, alias: str) -> None: ...
    def detach_catalog(self, alias: str) -> None: ...
    def detach_table(self, alias: str) -> None: ...
    def create_temp_table(self, ident: str, source: PyTableSource, replace: bool) -> Table: ...
    def current_catalog(self) -> Catalog | None: ...
    def current_namespace(self) -> PyIdentifier | None: ...
    def get_catalog(self, ident: str) -> Catalog: ...
    def get_table(self, ident: PyIdentifier) -> Table: ...
    def has_catalog(self, ident: str) -> bool: ...
    def has_table(self, ident: PyIdentifier) -> bool: ...
    def list_catalogs(self, pattern: str | None = None) -> list[str]: ...
    def list_tables(self, pattern: str | None = None) -> list[PyIdentifier]: ...
    def set_catalog(self, ident: str | None) -> None: ...
    def set_namespace(self, ident: PyIdentifier | None) -> None: ...
    def attach_function(self, function: UDF, alias: str | None = None) -> None: ...
    def detach_function(self, alias: str) -> None: ...

class InProgressShuffleCache:
    @staticmethod
    def try_new(
        num_partitions: int,
        dirs: list[str],
        node_id: str,
        shuffle_stage_id: int,
        target_filesize: int,
        compression: str | None = None,
        partition_by: list[PyExpr] | None = None,
    ) -> InProgressShuffleCache: ...
    async def push_partitions(self, input_partitions: list[PyMicroPartition]) -> None: ...
    async def close(self) -> ShuffleCache: ...

class ShuffleCache:
    def schema(self) -> PySchema: ...
    def file_paths_for_partition(self, partition_idx: int) -> list[str]: ...
    def bytes_per_file_for_partition(self, partition_idx: int) -> list[int]: ...
    def rows_per_partition(self) -> list[int]: ...
    def bytes_per_partition(self) -> list[int]: ...
    def clear_partition(self, partition_idx: int) -> None: ...
    def clear_directories(self) -> None: ...

class FlightServerConnectionHandle:
    def shutdown(self) -> None: ...
    def port(self) -> int: ...

def start_flight_server(
    shuffle_cache: ShuffleCache,
    ip: str,
) -> FlightServerConnectionHandle: ...

class FlightClientManager:
    def __init__(self, addresses: list[str], num_parallel_fetches: int, schema: PySchema): ...
    async def fetch_partition(self, partition: int) -> PyMicroPartition: ...

def cli(args: list[str]) -> None: ...

class PyScalarFunction:
    def __call__(self, *args: PyExpr, **kwargs: PyExpr) -> PyExpr: ...

def get_function_from_registry(name: str) -> PyScalarFunction: ...



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.datatype import DataType
from tests.utils import sort_arrow_table


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_distinct_with_nulls(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "id": [1, None, None, None],
            "values": ["a1", "b1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    result_df = daft_df.distinct()

    expected = {
        "id": [1, None, None],
        "values": ["a1", "b1", "c1"],
    }
    assert sort_arrow_table(pa.Table.from_pydict(result_df.to_pydict()), "values") == sort_arrow_table(
        pa.Table.from_pydict(expected), "values"
    )

    # Test unique alias.
    result_df = daft_df.unique()
    assert sort_arrow_table(pa.Table.from_pydict(result_df.to_pydict()), "values") == sort_arrow_table(
        pa.Table.from_pydict(expected), "values"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
def test_distinct_with_all_nulls(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "id": [None, None, None, None],
            "values": ["a1", "b1", "b1", "c1"],
        },
        repartition=repartition_nparts,
    )
    result_df = daft_df.select(daft_df["id"].cast(DataType.int64()), daft_df["values"]).distinct()

    expected = {
        "id": [None, None, None],
        "values": ["a1", "b1", "c1"],
    }
    assert sort_arrow_table(pa.Table.from_pydict(result_df.to_pydict()), "values") == sort_arrow_table(
        pa.Table.from_pydict(expected), "values"
    )

    # Test unique alias.
    result_df = daft_df.select(daft_df["id"].cast(DataType.int64()), daft_df["values"]).unique()
    assert sort_arrow_table(pa.Table.from_pydict(result_df.to_pydict()), "values") == sort_arrow_table(
        pa.Table.from_pydict(expected), "values"
    )


@pytest.mark.parametrize("repartition_nparts", [1, 2])
def test_distinct_with_empty(make_df, repartition_nparts, with_morsel_size):
    daft_df = make_df(
        {
            "id": [1],
            "values": ["a1"],
        },
        repartition=repartition_nparts,
    )
    result_df = daft_df.where(daft_df["id"] != 1).distinct()
    result_df.collect()

    resultset = result_df.to_pydict()
    assert len(resultset["id"]) == 0
    assert len(resultset["values"]) == 0

    # Test unique alias.
    result_df = daft_df.where(daft_df["id"] != 1).unique()
    result_df.collect()
    resultset = result_df.to_pydict()
    assert len(resultset["id"]) == 0
    assert len(resultset["values"]) == 0



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import Any

import pytest

from daft.expressions import (
    Expression,
    col,
    list_,
    lit,
)
from daft.expressions.visitor import PredicateVisitor
from daft.logical.schema import DataType
from daft.series import Series


@dataclass
class Trace:
    """Trace tells us which method was called."""

    attr: str
    args: list[Any]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Trace):
            return False
        if self.attr != other.attr:
            return False

        # recursive equals using the PyExpr._eq method
        def eq(obj1: object, obj2: object) -> bool:
            if isinstance(obj1, Expression) and isinstance(obj2, Expression):
                return obj1._expr._eq(obj2._expr)
            elif isinstance(obj1, list) and isinstance(obj2, list):
                if len(obj1) != len(obj2):
                    return False
                for lc, rc in zip(obj1, obj2):
                    if not eq(lc, rc):
                        return False
                return True
            else:
                return obj1 == obj2

        return eq(self.args, other.args)


class TracingVisitor(PredicateVisitor[list[Trace]]):
    """TracingVisitor accumulates a callstack so we can verify how the rust PyVisitor dispatches.

    This is factored intentionally to exemplify a stateless tree fold, and I've made the
    accumulator explicit. You could implement this much more concisely by making the visitor
    stateful with a `_traces: List[Trace]` field, but again this example is showing off
    functional style folding with dispatch on the rust side bridging into this visitor.

    I've also left the scoped context parameter out of the ExpressionVisitor interface because
    it's unlikely anyone would actually care for it, and anyone who does is already capable of
    managing their own state stack, this visitor is for the people.
    """

    def visit_col(self, name: str) -> list[Trace]:
        trace = Trace("visit_col", [name])
        acc = []
        acc += [trace]
        return acc

    def visit_lit(self, value: Any) -> list[Trace]:
        trace = Trace("visit_lit", [value])
        acc = []
        acc += [trace]
        return acc

    def visit_alias(self, expr: Expression, alias: str) -> list[Trace]:
        trace = Trace("visit_alias", [expr, alias])
        acc = []
        acc += self.visit(expr)
        acc += [trace]
        return acc

    def visit_cast(self, expr: Expression, dtype: DataType) -> list[Trace]:
        trace = Trace("visit_cast", [expr, dtype])
        acc = []
        acc += self.visit(expr)
        acc += [trace]
        return acc

    def visit_list(self, items: list[Expression]) -> list[Trace]:
        trace = Trace("visit_list", [items])
        acc = []
        for item in items:
            acc += self.visit(item)
        acc += [trace]
        return acc

    def visit_and(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_and", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_or(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_or", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_not(self, expr: Expression) -> list[Trace]:
        trace = Trace("visit_not", [expr])
        acc = []
        acc += self.visit(expr)
        acc += [trace]
        return acc

    def visit_equal(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_equal", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_not_equal(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_not_equal", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_less_than(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_less_than", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_less_than_or_equal(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_less_than_or_equal", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_greater_than(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_greater_than", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_greater_than_or_equal(self, left: Expression, right: Expression) -> list[Trace]:
        trace = Trace("visit_greater_than_or_equal", [left, right])
        acc = []
        acc += self.visit(left)
        acc += self.visit(right)
        acc += [trace]
        return acc

    def visit_between(self, expr: Expression, lower: Expression, upper: Expression) -> list[Trace]:
        trace = Trace("visit_between", [expr, lower, upper])
        acc = []
        acc += self.visit(expr)
        acc += self.visit(lower)
        acc += self.visit(upper)
        acc += [trace]
        return acc

    def visit_is_in(self, expr: Expression, items: list[Expression]) -> list[Trace]:
        trace = Trace("visit_is_in", [expr, items])
        acc = []
        acc += self.visit(expr)
        for item in items:
            acc += self.visit(item)
        acc += [trace]
        return acc

    def visit_is_null(self, expr: Expression) -> list[Trace]:
        trace = Trace("visit_is_null", [expr])
        acc = []
        acc += self.visit(expr)
        acc += [trace]
        return acc

    def visit_not_null(self, expr: Expression) -> list[Trace]:
        trace = Trace("visit_not_null", [expr])
        acc = []
        acc += self.visit(expr)
        acc += [trace]
        return acc

    def visit_function(self, name: str, args: list[Expression]) -> list[Trace]:
        trace = Trace("visit_function", [name, args])
        acc = []
        for arg in args:
            acc += self.visit(arg)
        acc += [trace]
        return acc


# tracing is stateless and therefore effectively just a function module, hence the singleton
TRACING = TracingVisitor()


def trace(expr: Expression) -> list[Trace]:
    return TRACING.visit(expr)


def test_visit_col():
    exp_1 = col("abc")
    assert trace(exp_1) == [Trace("visit_col", ["abc"])]


def test_visit_lit():
    # Test null literal
    lit_1 = lit(None)
    assert trace(lit_1) == [Trace("visit_lit", [None])]

    # Test boolean literals
    lit_2 = lit(True)
    lit_3 = lit(False)
    assert trace(lit_2) == [Trace("visit_lit", [True])]
    assert trace(lit_3) == [Trace("visit_lit", [False])]

    # Test string literal
    lit_4 = lit("hello")
    assert trace(lit_4) == [Trace("visit_lit", ["hello"])]

    # integer literal (always int64, casts are tested elsewhere)
    lit_5 = lit(1)
    assert trace(lit_5) == [Trace("visit_lit", [1])]

    # bytes literal
    lit_6 = lit(b"hello")
    assert trace(lit_6) == [Trace("visit_lit", [b"hello"])]

    # float literal
    lit_7 = lit(1.5)
    assert trace(lit_7) == [Trace("visit_lit", [1.5])]

    # decimal literal
    lit_8 = lit(Decimal("1.5"))
    assert trace(lit_8) == [Trace("visit_lit", [Decimal("1.5")])]

    # series literal (can't compare series directly, so manual compare..)
    lit_9 = lit(Series.from_pylist([1, 2, 3]))
    lit_9_trace = trace(lit_9)
    assert len(lit_9_trace) == 1
    assert len(lit_9_trace[0].args) == 1
    assert lit_9_trace[0].attr == "visit_lit"
    assert lit_9_trace[0].args[0].to_pylist() == [[1, 2, 3]]

    # struct literal
    lit_10 = lit({"a": 1, "b": 2})
    assert trace(lit_10) == [Trace("visit_lit", [{"a": 1, "b": 2}])]


@pytest.mark.parametrize(
    "data_type",
    [
        DataType.null(),
        DataType.bool(),
        # numbers
        DataType.int8(),
        DataType.int16(),
        DataType.int32(),
        DataType.int64(),
        DataType.uint8(),
        DataType.uint16(),
        DataType.uint32(),
        DataType.uint64(),
        DataType.decimal128(10, 2),
        DataType.float32(),
        DataType.float64(),
        # string
        DataType.string(),
        # datetime
        DataType.timestamp("ns"),
        DataType.date(),
        DataType.time("ns"),
        DataType.duration("ns"),
        DataType.interval(),
        # binary
        DataType.binary(),
        DataType.embedding(DataType.int64(), 512),
        # collections
        DataType.list(DataType.int64()),
        DataType.fixed_size_list(DataType.int64(), size=10),
        DataType.map(DataType.string(), DataType.int64()),
        DataType.struct({"a": DataType.int64(), "b": DataType.int64()}),
        # tensor
        DataType.tensor(DataType.int64()),
        DataType.tensor(DataType.int64(), shape=(2, 3)),
        DataType.sparse_tensor(DataType.int64()),
        DataType.sparse_tensor(DataType.int64(), shape=(2, 3)),
    ],
)
def test_visit_cast(data_type):
    exp = col("x").cast(data_type)
    assert trace(exp) == [Trace("visit_col", ["x"]), Trace("visit_cast", [col("x"), data_type])]


def test_visit_list():
    exp = list_(lit(1), lit(2), lit(3))
    assert trace(exp) == [
        Trace("visit_lit", [1]),
        Trace("visit_lit", [2]),
        Trace("visit_lit", [3]),
        Trace("visit_list", [[lit(1), lit(2), lit(3)]]),
    ]


def test_visit_equal():
    exp_0 = col("x")
    exp_1 = exp_0 == 1
    assert trace(exp_1) == [Trace("visit_col", ["x"]), Trace("visit_lit", [1]), Trace("visit_equal", [exp_0, lit(1)])]


def test_visit_not_equal():
    exp_0 = col("x")
    exp_1 = exp_0 != 1
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_not_equal", [exp_0, lit(1)]),
    ]


def test_visit_less_than():
    exp_0 = col("x")
    exp_1 = exp_0 < 1
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_less_than", [exp_0, lit(1)]),
    ]


def test_visit_less_than_or_equal():
    exp_0 = col("x")
    exp_1 = exp_0 <= 1
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_less_than_or_equal", [exp_0, lit(1)]),
    ]


def test_visit_greater_than():
    exp_0 = col("x")
    exp_1 = exp_0 > 1
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_greater_than", [exp_0, lit(1)]),
    ]


def test_visit_greater_than_or_equal():
    exp_0 = col("x")
    exp_1 = exp_0 >= 1
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_greater_than_or_equal", [exp_0, lit(1)]),
    ]


def test_visit_between():
    exp_0 = col("x")
    exp_1 = exp_0.between(1, 2)
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_lit", [2]),
        Trace("visit_between", [exp_0, lit(1), lit(2)]),
    ]


def test_visit_is_in():
    exp_0 = col("x")
    exp_1 = exp_0.is_in([1, 2, 3])
    assert trace(exp_1) == [
        Trace("visit_col", ["x"]),
        Trace("visit_lit", [1]),
        Trace("visit_lit", [2]),
        Trace("visit_lit", [3]),
        Trace("visit_is_in", [exp_0, [lit(1), lit(2), lit(3)]]),
    ]


def test_visit_is_null():
    exp_0 = col("x")
    exp_1 = exp_0.is_null()
    assert trace(exp_1) == [Trace("visit_col", ["x"]), Trace("visit_is_null", [exp_0])]


def test_visit_not_null():
    exp_0 = col("x")
    exp_1 = exp_0.not_null()
    assert trace(exp_1) == [Trace("visit_col", ["x"]), Trace("visit_not_null", [exp_0])]


@pytest.mark.parametrize(
    "attr",
    [
        "abs",
        "ceil",
        "floor",
        "sqrt",
        "exp",
        "log10",
        "sin",
        "cos",
        "tan",
        "arcsin",
        "arccos",
        "arctan",
        "sinh",
        "cosh",
        "tanh",
        "degrees",
        "radians",
        "sign",
    ],
)
def test_visit_functions0(attr):
    exp_0 = col("x")
    assert trace(getattr(exp_0, attr)()) == [
        Trace("visit_col", ["x"]),
        Trace("visit_function", [attr, [exp_0]]),
    ]



"""WARNING! THIS PACKAGE IS INTERNAL AND IS SUBJECT TO CHANGE."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

import boto3
import botocore
from botocore.exceptions import ClientError

from daft.catalog import Catalog, Identifier, NotFoundError, Properties, Table
from daft.datatype import DataType
from daft.logical.schema import Field, Schema

if TYPE_CHECKING:
    from mypy_boto3_glue import GlueClient
    from mypy_boto3_glue.type_defs import ColumnOutputTypeDef as GlueColumnInfo
    from mypy_boto3_glue.type_defs import TableTypeDef as GlueTableInfo
    from pyiceberg.table import Table as PyIcebergTable

    from daft.daft import IOConfig
    from daft.dataframe import DataFrame
    from daft.unity_catalog import UnityCatalogTable
else:
    GlueClient = Any
    GlueColumnInfo = Any
    GlueTableInfo = Any

Parameters = dict[str, Any]


def load_glue(
    name: str,
    region_name: str | None = None,
    api_version: str | None = None,
    use_ssl: bool | None = None,
    verify: bool | str | None = None,
    endpoint_url: str | None = None,
    aws_access_key_id: str | None = None,
    aws_secret_access_key: str | None = None,
    aws_session_token: str | None = None,
) -> GlueCatalog:
    """Creates a GlueCatalog with some give name and optional boto3 client configuration.

    Args:
        name (str): The name of the database to retrieve. For Hive compatibility, this should be all lowercase.
        region_name (str, optional): The AWS region to connect to.
        api_version (str, optional): The API version to use for the AWS Glue service.
        use_ssl (bool, optional): Whether to use SSL when connecting to AWS.
        verify (bool | str, optional): Whether to verify SSL certificates or a path to a CA bundle.
        endpoint_url (str, optional): Alternative endpoint URL to connect to.
        aws_access_key_id (str, optional): AWS access key ID for authentication.
        aws_secret_access_key (str, optional): AWS secret access key for authentication.
        aws_session_token (str, optional): AWS session token for temporary credentials.

    Returns:
        GlueCatalog: Catalog instance backed by an AWS Glue Database.
    """
    c = GlueCatalog.__new__(GlueCatalog)
    c._name = name
    options: dict[str, Any] = {}
    if region_name is not None:
        options["region_name"] = region_name
    if api_version is not None:
        options["api_version"] = api_version
    if use_ssl is not None:
        options["use_ssl"] = use_ssl
    if verify is not None:
        options["verify"] = verify
    if endpoint_url is not None:
        options["endpoint_url"] = endpoint_url
    if aws_access_key_id is not None:
        options["aws_access_key_id"] = aws_access_key_id
    if aws_secret_access_key is not None:
        options["aws_secret_access_key"] = aws_secret_access_key
    if aws_session_token is not None:
        options["aws_session_token"] = aws_session_token
    c._client = boto3.client("glue", **options)
    return c


class GlueCatalog(Catalog):
    """The GlueCatalog maps to an AWS Glue Database using a boto3 client."""

    _name: str
    _client: GlueClient

    # !! PATCH HERE TO PROVIDE CUSTOM GLUE TABLE IMPLEMENTATIONS !!
    _table_impls: list[type[GlueTable]] = []

    def __new__(cls) -> GlueCatalog:
        if cls is GlueCatalog:
            cls._table_impls = [
                GlueCsvTable,
                GlueParquetTable,
                GlueIcebergTable,
                GlueDeltaTable,
            ]
        return super().__new__(cls)

    def __init__(self) -> None:
        raise ValueError("GlueCatalog.__init__() not supported!")

    @staticmethod
    def from_client(name: str, client: GlueClient) -> GlueCatalog:
        """Creates a GlueCatalog using the given boto3 client."""
        c = GlueCatalog.__new__(GlueCatalog)
        c._name = name
        c._client = client
        return c

    @staticmethod
    def from_session(name: str, session: boto3.Session | botocore.session.Session) -> GlueCatalog:
        """Creates a GlueCatalog using the provided boto3 or botocore session to get a glue client.

        Args:
            name: Name of the catalog
            session: A boto3.Session or botocore.session.Session object

        Returns:
            A configured GlueCatalog instance
        """
        c = GlueCatalog.__new__(GlueCatalog)
        c._name = name
        if isinstance(session, boto3.Session):
            c._client = session.client("glue")
        elif isinstance(session, botocore.session.Session):
            c._client = session.create_client("glue")
        else:
            raise TypeError(f"Expected boto3.Session or botocore.session.Session, got {type(session).__name__}")

        return c

    @property
    def name(self) -> str:
        return self._name

    def _create_namespace(self, identifier: Identifier) -> None:
        try:
            self._client.create_database(
                DatabaseInput={
                    "Name": str(identifier),
                }
            )
        except self._client.exceptions.AlreadyExistsException:
            raise ValueError(f"Namespace {identifier} already exists")

    def _create_table(
        self,
        identifier: Identifier,
        schema: Schema,
        properties: Properties | None = None,
    ) -> Table:
        # Table creation implementation will be added later
        raise NotImplementedError("Table creation not yet implemented")

    def _drop_namespace(self, identifier: Identifier) -> None:
        try:
            self._client.delete_database(Name=str(identifier))
        except self._client.exceptions.EntityNotFoundException:
            raise NotFoundError(f"Namespace {identifier} not found")

    def _drop_table(self, identifier: Identifier) -> None:
        if len(identifier) != 2:
            raise ValueError(f"Expected identifier with form `<database_name>.<table_name>` but found {identifier}")
        database_name: str = str(identifier[0])
        name: str = str(identifier[1])

        try:
            self._client.delete_table(DatabaseName=database_name, Name=name)
        except self._client.exceptions.EntityNotFoundException:
            raise NotFoundError(f"Table {identifier} not found")

    def _get_table(self, identifier: Identifier) -> Table:
        if len(identifier) != 2:
            raise ValueError(f"Expected identifier with form `<database_name>.<table_name>` but found {identifier}")
        database_name: str = str(identifier[0])
        name: str = str(identifier[1])

        try:
            res = self._client.get_table(DatabaseName=database_name, Name=name)
            tbl = res["Table"]
            for impl in self._table_impls:
                try:
                    return impl.from_table_info(self, tbl)
                except ValueError:
                    pass
            classification = tbl["Parameters"].get("classification", "unknown")
            table_type = tbl["Parameters"].get("table_type", "unknown")
            raise ValueError(
                f"No supported table implementation for Table with classification='{classification}' and table_type='{table_type}'."
            )
        except self._client.exceptions.EntityNotFoundException:
            raise NotFoundError(f"Table {identifier} not found")

    def _has_table(self, identifier: Identifier) -> bool:
        try:
            self._get_table(identifier)
            return True
        except self._client.exceptions.EntityNotFoundException:
            return False

    def _has_namespace(self, identifier: Identifier) -> bool:
        try:
            self._client.get_database(Name=str(identifier))
            return True
        except self._client.exceptions.EntityNotFoundException:
            return False

    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        if pattern is not None:
            # Glue may add some kind of pattern to get_databases, then we can use their scheme.
            # Rather than making something up now which could later be broken.
            raise ValueError("GlueCatalog does not support using pattern filters with get_databases.")

        try:
            req = {}  # type: ignore
            namespaces = []
            while True:
                res = self._client.get_databases(**req)
                for database in res["DatabaseList"]:
                    namespaces.append(Identifier(database["Name"]))
                if next_token := res.get("NextToken"):
                    req["NextToken"] = next_token
                else:
                    break
            return namespaces
        except ClientError as e:
            raise ValueError("Failed to list namespaces.") from e

    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        if pattern is None:
            raise ValueError("GlueCatalog requires the pattern to contain a namespace.")

        req = {}
        if "." in pattern:
            database_name, expression = pattern.split(".", 1)
            req["DatabaseName"] = database_name
            req["Expression"] = expression
        else:
            req["DatabaseName"] = pattern

        try:
            tables = []
            while True:
                res = self._client.get_tables(**req)
                for table in res["TableList"]:
                    tables.append(Identifier(table["DatabaseName"], table["Name"]))
                if next_token := res.get("NextToken"):
                    req["NextToken"] = next_token
                else:
                    break
            return tables
        except ClientError as e:
            raise ValueError("Failed to list tables.") from e


class GlueTable(Table, ABC):
    """GlueTable is the base class for various Glue table formats."""

    _catalog: GlueCatalog
    _table: GlueTableInfo

    def __init__(self, catalog: GlueCatalog, table: GlueTableInfo):
        self._catalog = catalog
        self._table = table

    @classmethod
    @abstractmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        """Returns a GlueTable if the GlueTableInfo matches this implementation's requirements, otherwise raise a ValueError."""

    @property
    def name(self) -> str:
        return self._table["Name"]

    def schema(self) -> Schema:
        return self.read().schema()

    def __repr__(self) -> str:
        import json

        return json.dumps(self._table, indent=4, default=str)


class GlueCsvTable(GlueTable):
    """GlueCsvTable is for Glue classification='CSV' where we have delimited files under a common S3 prefix."""

    _path: str
    _schema: Schema
    _has_headers: bool = True
    _delimiter: str = ","
    _io_config: IOConfig | None = None
    _hive_partitioning: bool = False
    _hive_partitioning_cols: list[str] = []

    def __init__(self) -> None:
        raise ValueError("GlueCsvTable.__init__() not supported!")

    @classmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        # validate parameters information
        parameters: Parameters = table.get("Parameters", {})

        # check 'csv'
        classification = parameters.get("classification")
        if classification is None:
            raise ValueError("GlueTableInfo is missing the required parameter 'classification'.")
        if classification.lower() != "csv":
            raise ValueError(f"GlueTableInfo had classification {classification}, but expected 'CSV'")

        t = GlueCsvTable.__new__(GlueCsvTable)
        t._catalog = catalog
        t._table = table
        t._io_config = None  # todo

        # parse csv format information
        t._schema = _convert_glue_schema(table["StorageDescriptor"]["Columns"])
        t._path = table["StorageDescriptor"]["Location"]
        t._has_headers = parameters.get("skip.header.line.count", "0") == "1"
        t._delimiter = parameters.get("delimiter", ",")
        t._hive_partitioning = False
        t._hive_partitioning_cols = []

        return t

    def read(self, **options: Any) -> DataFrame:
        from daft.io._csv import read_csv

        return read_csv(
            path=self._path,
            infer_schema=False,
            schema={c.name: c.dtype for c in self._schema},
            has_headers=self._has_headers,
            delimiter=self._delimiter,
            double_quote=True,
            quote=None,
            escape_char=None,
            comment=None,
            allow_variable_columns=False,
            io_config=self._io_config,
            file_path_column=None,
            hive_partitioning=self._hive_partitioning,
        )

    def append(self, df: DataFrame, **options: Any) -> None:
        df.write_csv(
            root_dir=self._path,
            write_mode="append",
            partition_cols=(self._hive_partitioning_cols if self._hive_partitioning else None),  # type: ignore
        )

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        df.write_csv(
            root_dir=self._path,
            write_mode="overwrite",
            partition_cols=(self._hive_partitioning_cols if self._hive_partitioning else None),  # type: ignore
        )


class GlueParquetTable(GlueTable):
    """GlueParquetTable is for Glue classification='parquet' where we have parquet files under a common S3 prefix."""

    _path: str
    _schema: Schema
    _io_config: IOConfig | None = None
    _hive_partitioning: bool = False
    _hive_partitioning_cols: list[str] = []

    def __init__(self) -> None:
        raise ValueError("GlueParquetTable.__init__() not supported!")

    @classmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        # validate parameters information
        parameters: Parameters = table.get("Parameters", {})

        # check 'parquet'
        classification = parameters.get("classification")
        if classification is None:
            raise ValueError("GlueTableInfo is missing the required parameter 'classification'.")
        if classification.lower() != "parquet":
            raise ValueError(f"GlueTableInfo had classification {classification}, but expected 'parquet'")

        t = GlueParquetTable.__new__(GlueParquetTable)
        t._catalog = catalog
        t._table = table
        t._io_config = None  # todo

        # parse parquet format information
        t._schema = _convert_glue_schema(table["StorageDescriptor"]["Columns"])
        t._path = table["StorageDescriptor"]["Location"]

        return t

    def read(self, **options: Any) -> DataFrame:
        from daft.io._parquet import read_parquet

        return read_parquet(
            path=self._path,
            row_groups=None,
            infer_schema=False,
            schema={c.name: c.dtype for c in self._schema},
            io_config=self._io_config,
            file_path_column=None,
            hive_partitioning=self._hive_partitioning,
            coerce_int96_timestamp_unit=None,
            schema_hints=None,
        )

    def append(self, df: DataFrame, **options: Any) -> None:
        df.write_parquet(
            root_dir=self._path,
            compression="snappy",
            write_mode="append",
            partition_cols=(self._hive_partitioning_cols if self._hive_partitioning else None),  # type: ignore
            io_config=self._io_config,
        )

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        df.write_parquet(
            root_dir=self._path,
            compression="snappy",
            write_mode="overwrite",
            partition_cols=(self._hive_partitioning_cols if self._hive_partitioning else None),  # type: ignore
            io_config=self._io_config,
        )


class GlueIcebergTable(GlueTable):
    """GlueIcebergTable is for Glue table_type='ICEBERG'."""

    _io_config: IOConfig | None
    _pyiceberg_table: PyIcebergTable

    def __init__(self) -> None:
        raise ValueError("GlueIcebergTable.__init__() not supported!")

    @classmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        parameters: Parameters = table.get("Parameters", {})

        # verify we have table_type = "ICEBERG"
        table_type = parameters.get("table_type")
        if table_type is None:
            raise ValueError("GlueIcebegTable is missing the required 'table_type' parameter.")
        if table_type.lower() != "iceberg":
            raise ValueError(f"GlueIcebergTable had table_type {table_type}, but expected 'ICEBERG'")

        # create a pyiceberg catalog so we can write
        t = GlueIcebergTable.__new__(GlueIcebergTable)
        t._catalog = catalog
        t._table = table
        t._io_config = None  # todo
        t._pyiceberg_table = cls._create_pyiceberg_table(catalog, table)

        return t

    @classmethod
    def _create_pyiceberg_table(cls, catalog: GlueCatalog, table: GlueTableInfo) -> PyIcebergTable:
        from pyiceberg.catalog.glue import GlueCatalog as PyIcebergGlueCatalog

        # cannot create a pyiceberg GlueCatalog with custom client directly
        gc = PyIcebergGlueCatalog.__new__(PyIcebergGlueCatalog)
        gc.name = catalog.name
        gc.properties = {}
        gc.glue = catalog._client

        # the pyiceberg table will hold a ref to gc
        return gc._convert_glue_to_iceberg(table)

    def read(self, **options: Any) -> DataFrame:
        from daft.io._iceberg import read_iceberg

        return read_iceberg(
            table=self._pyiceberg_table, snapshot_id=options.get("snapshot_id"), io_config=self._io_config
        )

    def append(self, df: DataFrame, **options: Any) -> None:
        df.write_iceberg(self._pyiceberg_table, mode="append")

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        df.write_iceberg(self._pyiceberg_table, mode="overwrite")


class GlueDeltaTable(GlueTable):
    """GlueDeltaTable is for Glue table_type='ICEBERG'."""

    _io_config: IOConfig | None
    _unity_catalog_table: UnityCatalogTable

    def __init__(self) -> None:
        raise ValueError("GlueDeltaTable.__init__() not supported!")

    @classmethod
    def from_table_info(cls, catalog: GlueCatalog, table: GlueTableInfo) -> GlueTable:
        parameters: Parameters = table.get("Parameters", {})

        # verify we have table_type = "delta"
        table_type = parameters.get("table_type")
        if table_type is None:
            raise ValueError("GlueDeltaTable is missing the required 'table_type' parameter.")
        if table_type.lower() != "delta":
            raise ValueError(f"GlueDeltaTable had table_type {table_type}, but expected 'delta'")

        t = GlueDeltaTable.__new__(GlueDeltaTable)
        t._catalog = catalog
        t._table = table
        t._io_config = None  # todo
        t._unity_catalog_table = cls._create_unity_catalog_table(table, None)

        return t

    @classmethod
    def _create_unity_catalog_table(cls, table: GlueTableInfo, io_config: IOConfig | None) -> UnityCatalogTable:
        from unitycatalog.types import TableInfo

        from daft.unity_catalog import UnityCatalogTable

        table_info = TableInfo(
            # catalog_name=
            # columns=[],
            comment=None,
            created_at=int(table["CreateTime"].timestamp() * 1000),
            data_source_format="DELTA",
            schema_name=table["DatabaseName"],
        )

        table_uri = table["StorageDescriptor"]["Location"]

        return UnityCatalogTable(table_info, table_uri, io_config)

    def read(self, **options: Any) -> DataFrame:
        from daft.io._deltalake import read_deltalake

        return read_deltalake(
            table=self._unity_catalog_table,
            version=options.get("version"),
            io_config=self._io_config,
        )

    def append(self, df: DataFrame, **options: Any) -> None:
        raise NotImplementedError

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        raise NotImplementedError


def _convert_glue_schema(columns: list[GlueColumnInfo]) -> Schema:
    return Schema._from_fields([_convert_glue_column(column) for column in columns])


def _convert_glue_column(column: GlueColumnInfo) -> Field:
    return Field.create(column["Name"], _convert_glue_type(column["Type"]))


# TODO update to support more Hive types.
# https://cwiki.apache.org/confluence/display/hive/languagemanual+types
def _convert_glue_type(type: str) -> DataType:
    type = type.lower()
    if type == "boolean":
        return DataType.bool()
    elif type == "byte":
        return DataType.int8()
    elif type == "short":
        return DataType.int16()
    elif type == "integer":
        return DataType.int32()
    elif type == "long" or type == "bigint":
        return DataType.int64()
    elif type == "float":
        return DataType.float32()
    elif type == "double":
        return DataType.float64()
    elif type == "decimal":
        return DataType.decimal128(precision=38, scale=18)
    elif type == "string":
        return DataType.string()
    elif type == "timestamp":
        return DataType.timestamp(timeunit="us", timezone="UTC")
    elif type == "date":
        return DataType.date()
    else:
        raise ValueError(f"Unsupported Glue type: {type}")



from __future__ import annotations

import pytest

from daft.expressions import col, lit
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    ["expr", "data", "expected"],
    [
        (col("col").str.lpad(5, "."), ["foo", "abcdef", "quux"], ["..foo", "abcde", ".quux"]),
        (col("col").str.lpad(lit(5), lit("-")), ["foo", "abcdef", "quux"], ["--foo", "abcde", "-quux"]),
        (
            col("col").str.lpad(col("zeroes") + lit(5), col("emptystrings") + lit("-")),
            ["foo", "abcdef", "quux"],
            ["--foo", "abcde", "-quux"],
        ),
    ],
)
def test_series_utf8_lpad_broadcast_pattern(expr, data, expected) -> None:
    table = MicroPartition.from_pydict({"col": data, "emptystrings": ["", "", ""], "zeroes": [0, 0, 0]})
    result = table.eval_expression_list([expr])
    assert result.to_pydict() == {"col": expected}



from __future__ import annotations

from typing import Literal

import pytest

from daft import DataType, Series


def minhash_none(
    series: Series,
    num_hashes: int,
    ngram_size: int,
    seed: int | None,
    hash_function: Literal["murmurhash3", "xxhash", "sha1"] = "murmurhash3",
) -> list[list[int] | None]:
    if seed is None:
        return series.minhash(num_hashes, ngram_size, hash_function=hash_function).to_pylist()
    else:
        return series.minhash(num_hashes, ngram_size, seed, hash_function=hash_function).to_pylist()


test_series = Series.from_pylist(
    [
        "The quick brown fox",
        "The speedy orange fox",
        "The quick brown fox",
        "thisonlyhasonetokenohno",
        None,
        "This has more than four tokens, unlike the other strings",
        "!@# $%^&*() -` 1235 827 9387 216340",
        "This   has    excessive\t\n and   weird    whitespace",
        "",
        " spaces at start and end ",
        " ",
        None,
    ]
)


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash(num_hashes, ngram_size, seed, hash_function):
    minhash = minhash_none(test_series, num_hashes, ngram_size, seed, hash_function)
    assert minhash[4] is None and minhash[-1] is None
    for lst in minhash:
        if lst is not None:
            assert len(lst) == num_hashes
    for i in range(num_hashes):
        assert minhash[0][i] == minhash[2][i]
        if ngram_size > 1:
            assert minhash[0][i] != minhash[1][i]


@pytest.mark.parametrize(
    "num_hashes,ngram_size,seed,expected",
    [
        # Test with single hash, unigrams
        (
            1,
            1,
            1,
            [
                [1196831525],  # "The quick brown fox"
                [120174860],  # "The speedy orange fox"
                [1196831525],  # "The quick brown fox" - identical to first
                [2559787809],  # "thisonlyhasonetokenohno"
                None,  # None value
                [27473697],  # "This has more..."
                [441506281],  # "!@# $%^&*()..."
                [27473697],  # "This has excessive..."
                # [500470364],  # "" - empty string todo(andrewgazelka): fix empty string
                [4294967295],  # todo: this is different than previous impl ^
                [76461626],  # " spaces at..."
                [500470364],  # " " - just a space
                None,  # None value
            ],
        ),
        # Test with two hashes, bigrams
        (
            2,
            2,
            123,
            [
                [760527683, 1539127776],  # "The quick brown fox"
                [1704758042, 309185920],  # "The speedy orange fox"
                [760527683, 1539127776],  # "The quick brown fox" - identical to first
                [3763775515, 2389564536],  # "thisonlyhasonetokenohno"
                None,  # None value
                [437177734, 1262955240],  # "This has more..."
                [101182009, 511203536],  # "!@# $%^&*()..."
                [27545328, 189622288],  # "This has excessive..."
                # [2989311896, 1304790168],  # "" - empty string
                [4294967295, 4294967295],  # todo: this is different than previous impl ^
                [94241209, 101414440],  # " spaces at start and end "
                [531691842, 296683088],  # " " - just a space
                None,  # None value
            ],
        ),
    ],
)
def test_minhash_exact_values(num_hashes, ngram_size, seed, expected):
    result = minhash_none(test_series, num_hashes, ngram_size, seed)
    assert result == expected


@pytest.mark.parametrize("num_hashes", [0, -1, -100])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash_fails_nonpositive_num_hashes(num_hashes, ngram_size, seed, hash_function):
    with pytest.raises(ValueError, match="num_hashes must be positive"):
        minhash_none(test_series, num_hashes, ngram_size, seed, hash_function)


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [0, -1, -100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash_fails_nonpositive_ngram_size(num_hashes, ngram_size, seed, hash_function):
    with pytest.raises(ValueError, match="ngram_size must be positive"):
        minhash_none(test_series, num_hashes, ngram_size, seed, hash_function)


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash_empty_series(num_hashes, ngram_size, seed, hash_function):
    series = Series.from_pylist([]).cast(DataType.string())

    minhash = minhash_none(series, num_hashes, ngram_size, seed, hash_function)
    assert len(minhash) == 0


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash_seed_consistency(num_hashes, ngram_size, seed, hash_function):
    minhash1 = minhash_none(test_series, num_hashes, ngram_size, seed, hash_function)
    minhash2 = minhash_none(test_series, num_hashes, ngram_size, seed, hash_function)
    assert minhash1 == minhash2


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed_pair", [[1, 2], [1, 5], [None, 2], [123, 234]])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_minhash_seed_differences(num_hashes, ngram_size, seed_pair, hash_function):
    minhash1 = minhash_none(test_series, num_hashes, ngram_size, seed_pair[0], hash_function)
    minhash2 = minhash_none(test_series, num_hashes, ngram_size, seed_pair[1], hash_function)
    assert minhash1 != minhash2



from __future__ import annotations

from typing import Any

from daft.catalog import Catalog, Identifier, Properties, Table
from daft.daft import PyCatalog as _PyCatalog
from daft.daft import PyTable as _PyTable
from daft.dataframe import DataFrame
from daft.logical.builder import LogicalPlanBuilder
from daft.schema import Schema


class _RustCatalog(Catalog):
    """Shim to wrap PyCatalog and subclass Catalog.

    This should not be used directly, but instead be subclassed for each Rust catalog.
    """

    def __init__(self, inner: _PyCatalog):
        self.inner = inner

    @property
    def name(self) -> str:
        return self.inner.name()

    def _create_namespace(self, ident: Identifier) -> None:
        self.inner.create_namespace(ident._ident)

    def _create_table(self, ident: Identifier, schema: Schema, properties: Properties | None = None) -> Table:
        return self.inner.create_table(ident._ident, schema._schema)

    def _drop_namespace(self, ident: Identifier) -> None:
        self.inner.drop_namespace(ident._ident)

    def _drop_table(self, ident: Identifier) -> None:
        self.inner.drop_table(ident._ident)

    def _get_table(self, ident: Identifier) -> Table:
        return self.inner.get_table(ident._ident)

    def _has_namespace(self, ident: Identifier) -> bool:
        return self.inner.has_namespace(ident._ident)

    def _has_table(self, ident: Identifier) -> bool:
        return self.inner.has_table(ident._ident)

    def _list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        return [Identifier._from_pyidentifier(ident) for ident in self.inner.list_namespaces(pattern)]

    def _list_tables(self, pattern: str | None = None) -> list[Identifier]:
        return [Identifier._from_pyidentifier(ident) for ident in self.inner.list_tables(pattern)]


class _RustTable(Table):
    """Shim to wrap PyTable and subclass Table.

    This should not be used directly, but instead be subclassed for each Rust table.
    """

    def __init__(self, inner: _PyTable):
        self.inner = inner

    @property
    def name(self) -> str:
        return self.inner.name()

    def schema(self) -> Schema:
        return Schema._from_pyschema(self.inner.schema())

    def read(self, **options: Any) -> DataFrame:
        return DataFrame(LogicalPlanBuilder(self.inner.to_logical_plan()))

    def append(self, df: DataFrame, **options: Any) -> None:
        self.inner.append(df._builder._builder, **options)

    def overwrite(self, df: DataFrame, **options: Any) -> None:
        self.inner.overwrite(df._builder._builder, **options)


class View(_RustTable):
    pass


class MemoryCatalog(_RustCatalog):
    @staticmethod
    def _new(name: str) -> Catalog:
        return _PyCatalog.new_memory_catalog(name)


class MemoryTable(_RustTable):
    @staticmethod
    def _new(name: str, schema: Schema) -> Table:
        return _PyTable.new_memory_table(name, schema._schema)



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Callable

from daft.daft import PyPartitionField, PyPushdowns, PyRecordBatch, ScanOperatorHandle, ScanTask
from daft.dataframe import DataFrame
from daft.io.scan import ScanOperator
from daft.logical.builder import LogicalPlanBuilder
from daft.logical.schema import Schema

if TYPE_CHECKING:
    from daft.recordbatch.recordbatch import RecordBatch


def _generator_factory_function(func: Callable[[], Iterator["RecordBatch"]]) -> Iterator[PyRecordBatch]:
    for table in func():
        yield table._recordbatch


def read_generator(
    generators: Iterator[Callable[[], Iterator["RecordBatch"]]],
    schema: Schema,
) -> DataFrame:
    """Create a DataFrame from a generator function.

    Args:
        generator (Callable[[int, Any], Iterator[RecordBatch]]): a generator function that generates data
        num_partitions (int): the number of partitions to generate
        schema (Schema): the schema of the generated data
        generator_args (Any): additional arguments to pass to the generator

    Returns:
        DataFrame: a DataFrame containing the generated data

    Examples:
        >>> import daft
        >>> from daft.io._generator import read_generator
        >>> from daft.recordbatch.recordbatch import RecordBatch
        >>> from functools import partial
        >>>
        >>> # Set runner to Ray for distributed processing
        >>> daft.context.set_runner_ray()
        >>>
        >>> # Helper function to generate data for each partition
        >>> def generate(num_rows: int):
        ...     data = {"ints": [i for i in range(num_rows)]}
        ...     yield RecordBatch.from_pydict(data)
        >>>
        >>> # Generator function that yields partial functions for each partition
        >>> def generator(num_partitions: int):
        ...     for i in range(num_partitions):
        ...         yield partial(generate, 100)
        >>>
        >>> # Create DataFrame using read_generator and repartition the data
        >>> df = (
        ...     read_generator(
        ...         generator(num_partitions=100),
        ...         daft.Schema._from_field_name_and_types([("ints", daft.DataType.uint64())]),
        ...     )
        ...     .repartition(100, "ints")
        ...     .collect()
        ... )
    """
    generator_scan_operator = GeneratorScanOperator(
        generators=generators,
        schema=schema,
    )
    handle = ScanOperatorHandle.from_python_scan_operator(generator_scan_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)


class GeneratorScanOperator(ScanOperator):
    def __init__(
        self,
        generators: Iterator[Callable[[], Iterator["RecordBatch"]]],
        schema: Schema,
    ):
        self._generators = list(generators)
        self._schema = schema

    def name(self) -> str:
        return self.display_name()

    def display_name(self) -> str:
        return "GeneratorScanOperator"

    def schema(self) -> Schema:
        return self._schema

    def partitioning_keys(self) -> list[PyPartitionField]:
        return []

    def can_absorb_filter(self) -> bool:
        return False

    def can_absorb_limit(self) -> bool:
        return False

    def can_absorb_select(self) -> bool:
        return False

    def multiline_display(self) -> list[str]:
        return [
            self.display_name(),
            f"Schema = {self.schema()}",
        ]

    def to_scan_tasks(self, pushdowns: PyPushdowns) -> Iterator[ScanTask]:
        for generator in self._generators:
            yield ScanTask.python_factory_func_scan_task(
                module=_generator_factory_function.__module__,
                func_name=_generator_factory_function.__name__,
                func_args=(generator,),
                schema=self.schema()._schema,
                num_rows=None,
                size_bytes=None,
                pushdowns=pushdowns,
                stats=None,
            )



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

import inspect
import warnings
from typing import Optional

import daft
from daft.api_annotations import PublicAPI
from daft.context import get_context
from daft.daft import LogicalPlanBuilder as _PyLogicalPlanBuilder
from daft.daft import PySqlCatalog as _PySqlCatalog
from daft.daft import sql_exec as _sql_exec
from daft.daft import sql_expr as _sql_expr
from daft.dataframe import DataFrame
from daft.exceptions import DaftCoreException
from daft.expressions import Expression
from daft.logical.builder import LogicalPlanBuilder


class SQLCatalog:
    """SQLCatalog is a simple map from table names to dataframes used in query planning.

    EXPERIMENTAL: This features is early in development and will change.
    """

    _catalog: _PySqlCatalog = None  # type: ignore

    def __post_init__(self) -> None:
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.6.0; please use `Catalog.from_pydict()`.",
            DeprecationWarning,
            stacklevel=2,
        )

    def __init__(self, tables: dict[str, DataFrame]) -> None:
        """Create a new SQLCatalog from a dictionary of table names to dataframes."""
        self._catalog = _PySqlCatalog.new()
        for name, df in tables.items():
            self._catalog.register_table(name, df._get_current_builder()._builder)

    def __str__(self) -> str:
        return str(self._catalog)

    def register_table(self, name: str, df: DataFrame) -> None:
        self._catalog.register_table(name, df._get_current_builder()._builder)

    def _copy_from(self, other: "SQLCatalog") -> None:
        self._catalog.copy_from(other._catalog)


@PublicAPI
def sql_expr(sql: str) -> Expression:
    """Parses a SQL string into a Daft Expression.

    This function allows you to create Daft Expressions from SQL snippets, which can then be used
    in Daft operations or combined with other Daft Expressions.

    Args:
        sql (str): A SQL string to be parsed into a Daft Expression.

    Returns:
        Expression: A Daft Expression representing the parsed SQL.

    Examples:
        Create a simple SQL expression:

        >>> import daft
        >>> expr = daft.sql_expr("1 + 2")
        >>> print(expr)
        lit(1) + lit(2)

        Use SQL expression in a Daft DataFrame operation:

        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> df = df.with_column("c", daft.sql_expr("a + b"))
        >>> df.show()
        ╭───────┬───────┬───────╮
        │ a     ┆ b     ┆ c     │
        │ ---   ┆ ---   ┆ ---   │
        │ Int64 ┆ Int64 ┆ Int64 │
        ╞═══════╪═══════╪═══════╡
        │ 1     ┆ 4     ┆ 5     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 2     ┆ 5     ┆ 7     │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌┤
        │ 3     ┆ 6     ┆ 9     │
        ╰───────┴───────┴───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        `daft.sql_expr` is also called automatically for you in some DataFrame operations such as filters:

        >>> df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
        >>> result = df.where("x < 3 AND y > 4")
        >>> result.show()
        ╭───────┬───────╮
        │ x     ┆ y     │
        │ ---   ┆ ---   │
        │ Int64 ┆ Int64 │
        ╞═══════╪═══════╡
        │ 2     ┆ 5     │
        ╰───────┴───────╯
        <BLANKLINE>
        (Showing first 1 of 1 rows)
    """
    return Expression._from_pyexpr(_sql_expr(sql))


@PublicAPI
def sql(
    sql: str,
    catalog: Optional[SQLCatalog] = None,
    register_globals: bool = True,
    **bindings: DataFrame,
) -> DataFrame:
    """Run a SQL query, returning the results as a DataFrame.

    Args:
        sql (str): SQL query to execute
        catalog (SQLCatalog, optional): Catalog of tables to use in the query.
            Defaults to None, in which case a catalog will be built from variables
            in the callers scope.
        register_globals (bool, optional): Whether to incorporate global
            variables into the supplied catalog, in which case a copy of the
            catalog will be made and the original not modified. Defaults to True.
        **bindings: (DataFrame): Additional DataFrame bindings (CTEs) to use for this query.

    Returns:
        DataFrame: Dataframe containing the results of the query

    Warning:
        This features is early in development and will likely experience API changes.

    Examples:
        A simple example joining 2 dataframes together using a SQL statement, relying on Daft to detect the names of
        SQL tables using their corresponding Python variable names.

        >>> import daft
        >>>
        >>> df1 = daft.from_pydict({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})
        >>> df2 = daft.from_pydict({"a": [1, 2, 3], "c": ["daft", None, None]})
        >>>
        >>> # Daft automatically detects `df1` and `df2` from your Python global namespace
        >>> result_df = daft.sql("SELECT * FROM df1 JOIN df2 ON df1.a = df2.a")
        >>> result_df.show()
        ╭───────┬──────┬──────╮
        │ a     ┆ b    ┆ c    │
        │ ---   ┆ ---  ┆ ---  │
        │ Int64 ┆ Utf8 ┆ Utf8 │
        ╞═══════╪══════╪══════╡
        │ 1     ┆ foo  ┆ daft │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 2     ┆ bar  ┆ None │
        ├╌╌╌╌╌╌╌┼╌╌╌╌╌╌┼╌╌╌╌╌╌┤
        │ 3     ┆ baz  ┆ None │
        ╰───────┴──────┴──────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)

        A more complex example using CTE bindings to create a named subquery (DataFrame) called `"my_df"`, which can then be referenced from inside your SQL statement.

        >>> import daft
        >>> from daft.sql import SQLCatalog
        >>>
        >>> df = daft.from_pydict({"a": [1, 2, 3], "b": ["foo", "bar", "baz"]})
        >>>
        >>> # Register dataframes as table expressions using a python dictionary.
        >>> bindings = {"my_df": df}
        >>>
        >>> daft.sql("SELECT a FROM my_df", **bindings).show()
        ╭───────╮
        │ a     │
        │ ---   │
        │ Int64 │
        ╞═══════╡
        │ 1     │
        ├╌╌╌╌╌╌╌┤
        │ 2     │
        ├╌╌╌╌╌╌╌┤
        │ 3     │
        ╰───────╯
        <BLANKLINE>
        (Showing first 3 of 3 rows)
    """
    # This the CTE bindings map which is built in the order globals->catalog->ctes.
    py_ctes: dict[str, _PyLogicalPlanBuilder] = {}

    # 1. Add all python DataFrame variables which are in scope.
    if register_globals:
        try:
            # Caller is back from func, analytics, annotation
            caller_frame = inspect.currentframe().f_back.f_back.f_back  # type: ignore
            caller_vars = {**caller_frame.f_globals, **caller_frame.f_locals}  # type: ignore
        except AttributeError as exc:
            # some interpreters might not implement currentframe; all reasonable
            # errors above should be AttributeError
            raise DaftCoreException(
                "Cannot get caller environment, please provide CTEs and set `register_globals=False`."
            ) from exc
        for alias, variable in caller_vars.items():
            if isinstance(variable, DataFrame):
                py_ctes[alias] = variable._builder._builder

    # 2. Add all SQLCatalog names for backwards compatibility.
    if catalog:
        warnings.warn(
            "The `catalog` argument is deprecated and will be removed in daft >= 0.6.0. Please use `ctes` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        py_ctes.update(catalog._catalog.to_pydict())

    # 3. Add explicit CTEs last so these can't be shadowed.
    for alias, df in bindings.items():
        py_ctes[alias] = df._builder._builder

    py_sess = daft.current_session()._session
    py_config = get_context().daft_planning_config
    py_object = _sql_exec(sql, py_sess, py_ctes, py_config)

    if py_object is None:
        # for backwards compatibility on the return type i.e. don't introduce nullability
        return DataFrame._from_pydict({})
    elif isinstance(py_object, _PyLogicalPlanBuilder):
        return DataFrame(LogicalPlanBuilder(py_object))
    else:
        raise ValueError(f"Unsupported return type from sql exec: {type(py_object)}")



from __future__ import annotations

import pandas as pd

from daft import DataType, col, lit
from tests.conftest import assert_df_equals


def test_literal_column(daft_df, service_requests_csv_pd_df):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.with_column("literal_col", lit(1))
    daft_pd_df = daft_df.to_pandas()
    service_requests_csv_pd_df["literal_col"] = 1
    service_requests_csv_pd_df["literal_col"] = service_requests_csv_pd_df["literal_col"].astype("int32")
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_literal_column_computation(daft_df, service_requests_csv_pd_df):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.with_column("literal_col", lit(1) + 1)
    daft_pd_df = daft_df.to_pandas()
    service_requests_csv_pd_df["literal_col"] = 1 + 1
    service_requests_csv_pd_df["literal_col"] = service_requests_csv_pd_df["literal_col"].astype("int32")
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_literal_column_aggregation(daft_df, service_requests_csv_pd_df):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.repartition(2).groupby("Borough").agg(col("Unique Key").sum())
    daft_df = daft_df.with_column("literal_col", lit(1) + 1)
    daft_pd_df = daft_df.to_pandas()
    service_requests_csv_pd_df = service_requests_csv_pd_df.groupby("Borough").agg({"Unique Key": "sum"})
    service_requests_csv_pd_df = service_requests_csv_pd_df.reset_index()[["Unique Key", "Borough"]]
    service_requests_csv_pd_df["literal_col"] = 1 + 1
    service_requests_csv_pd_df["literal_col"] = service_requests_csv_pd_df["literal_col"].astype("int32")
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df, sort_key="Borough")


def test_pyobj_literal_column(daft_df, service_requests_csv_pd_df):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.with_column("literal_col", lit({"foo": "bar"}))
    daft_pd_df = daft_df.to_pandas()
    service_requests_csv_pd_df["literal_col"] = pd.Series(
        {"foo": "bar"} for _ in range(len(service_requests_csv_pd_df))
    )
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)


def test_literal_column_computation_apply(daft_df, service_requests_csv_pd_df):
    """Creating a new column that is derived from (1 + other_column) and retrieving the top N results."""
    daft_df = daft_df.with_column(
        "literal_col", lit({"foo": "bar"}).apply(lambda d: d["foo"], return_dtype=DataType.string())
    )
    daft_pd_df = daft_df.to_pandas()
    service_requests_csv_pd_df["literal_col"] = "bar"
    assert_df_equals(daft_pd_df, service_requests_csv_pd_df)



# def test_apply_lambda
# def test_apply_module_func
# def test_apply_inline_func
# def test_apply_lambda_pyobj

from __future__ import annotations

import dataclasses

import daft
from daft import DataType


def add_1(x):
    return x + 1


def test_apply_module_func():
    df = daft.from_pydict({"a": [1, 2, 3]})
    df = df.with_column("a_plus_1", df["a"].apply(add_1, return_dtype=DataType.int32()))
    assert df.to_pydict() == {"a": [1, 2, 3], "a_plus_1": [2, 3, 4]}


def test_apply_lambda():
    df = daft.from_pydict({"a": [1, 2, 3]})
    df = df.with_column("a_plus_1", df["a"].apply(lambda x: x + 1, return_dtype=DataType.int32()))
    assert df.to_pydict() == {"a": [1, 2, 3], "a_plus_1": [2, 3, 4]}


def test_apply_inline_func():
    def inline_add_1(x):
        return x + 1

    df = daft.from_pydict({"a": [1, 2, 3]})
    df = df.with_column("a_plus_1", df["a"].apply(inline_add_1, return_dtype=DataType.int32()))
    assert df.to_pydict() == {"a": [1, 2, 3], "a_plus_1": [2, 3, 4]}


@dataclasses.dataclass
class MyObj:
    x: int


def test_apply_obj():
    df = daft.from_pydict({"obj": [MyObj(x=0), MyObj(x=0), MyObj(x=0)]})

    def inline_mutate_obj(obj):
        obj.x = 1
        return obj

    df = df.with_column("mut_obj", df["obj"].apply(inline_mutate_obj, return_dtype=DataType.python()))
    result = df.to_pydict()
    for mut_obj in result["mut_obj"]:
        assert mut_obj.x == 1
